/* *
 * Copyright 1993-2012 NVIDIA Corporation.  All rights reserved.
 *
 * Please refer to the NVIDIA end user license agreement (EULA) associated
 * with this source code for terms and conditions that govern your use of
 * this software. Any use, reproduction, disclosure, or distribution of
 * this software and related documentation outside the terms of the EULA
 * is strictly prohibited.
 */

/* *
 * Read an .AUT file and relabel it (replace action strings by numbers), and sort outgoing transitions per src state by label, v.1.0 by Anton Wijs, 2014
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <time.h>
#include <math.h>
#include <map>
#include "uthash.h"

static const unsigned int IOBUFFERSIZE = 7000;
static const unsigned int LABELSIZE = 7000;
static const unsigned int EMPTYSTATE = 0x7FFFFFFF;
static const unsigned int LAST = 0xFFFFFFFF;


// BIT MANIPULATION MACROS

#define MIN(a,b) \
   ({ __typeof__ (a) _a = (a); \
       __typeof__ (b) _b = (b); \
     _a < _b ? _a : _b; })

#define MAX(a,b) \
   ({ __typeof__ (a) _a = (a); \
       __typeof__ (b) _b = (b); \
     _a > _b ? _a : _b; })

// hashable struct for obtaining the final state IDs of the output (new) LTSs
typedef struct actsetnode {
	char act[LABELSIZE]; // key is action label
	unsigned int id; // new id for a label
	UT_hash_handle hh; // makes node hashable
} actsetnode;

// Global variables
unsigned int nstates, ntrans;
unsigned int *offsets, *transact, *transtgt;
// needed for output LTS
unsigned int *permutation, *tmparray1, *tmparray2;
// hashtable to assign new IDs to labels in new LTS
actsetnode *newactIDs;
unsigned int actIDcounter = 1;
// offset to tell the compare function which results to look at
unsigned int compare_offset = 0;

// compare function to sort the output LTS transitions
int compare (const void * a, const void * b)
{
	return ( transact[compare_offset+(*(int*)a)] - transact[compare_offset+(*(int*)b)] );
}

// HASHTABLE FUNCTIONS

// clean the action set
void clean_actset() {
	actsetnode *current, *tmp;
	
	HASH_ITER(hh, newactIDs, current, tmp) {
		HASH_DEL(newactIDs, current);
		free(current);
	}
	actIDcounter = 0;
}

// add an action to the action set and return its ID
unsigned int addact(char *act) {
	actsetnode *s;

	HASH_FIND_STR(newactIDs, act, s);
	if (s==NULL) {
		s = (actsetnode*) malloc(sizeof(actsetnode));
		strcpy(s->act, act);
		s->id = actIDcounter;
		HASH_ADD_STR(newactIDs, act, s);
		return actIDcounter++;
	}
	else {
		return s->id;
	}
}

/**
 * Host function that reads MDP file (of PRISM) and launches algorithm.
 */
int main(int argc, char** argv) {
	FILE *fp;
	// level of verbosity (1=print progress)
	int verbosity = 0;
	char stmp1[IOBUFFERSIZE], stmp2[IOBUFFERSIZE], fn[50], fn2[55];
	unsigned int i, j, k, size, max_brfactor;

	// Structures to store AUT input
	unsigned int src_state, tgt_state, offsets_cnt, trans_cnt, actid;
	char act[LABELSIZE];

	if (argc == 1) {
		fprintf(stderr, "ERROR: No input AUT given!\n");
		exit(1);
	}

	strcpy(fn, argv[1]);

	i = 2;
	while (i < argc) {
		if (!strcmp(argv[i],"-v")) {
			// store verbosity level
			verbosity = atoi(argv[i+1]);
			if (verbosity > 3) {
				verbosity = 3;
			}
			i += 2;
		}
	}

	fp = fopen(fn, "r");
	if (fp) {
		// Read the input
		fgets(stmp1, IOBUFFERSIZE, fp);

		char *sp1, *sp2;
		sp1 = strchr(stmp1, ',');
		sp1++;
		while ((*sp1) == ' ') {
			sp1++;
		}
		sp2 = sp1;
		while ((*sp2) != ' ' && (*sp2) != ',') {
			sp2++;
		}
		memcpy (stmp2, sp1, sp2-sp1);
		stmp2[sp2-sp1] = '\0';
		ntrans = (unsigned int) atoi(stmp2);
		fprintf(stdout, "nr of transitions in LTS: %d\n", ntrans);
		sp1 = strchr(sp2, ',');
		sp1++;
		while ((*sp1) == ' ') {
			sp1++;
		}
		sp2 = sp1;
		while ((*sp2) != ' ' && (*sp2) != ')') {
			sp2++;
		}
		memcpy (stmp2, sp1, sp2-sp1);
		stmp2[sp2-sp1] = '\0';
		nstates = (unsigned int) atoi(stmp2);
		fprintf(stdout, "nr of states in LTS: %d\n", nstates);

		// fill the arrays with the LTS transitions
		offsets = (unsigned int*) malloc(sizeof(unsigned int)*(nstates+1));
		transact = (unsigned int*) malloc(sizeof(unsigned int)*ntrans);
		transtgt = (unsigned int*) malloc(sizeof(unsigned int)*ntrans);
		offsets_cnt = 0; trans_cnt = 0;
		offsets[offsets_cnt++] = 0;
		j = 0;
		max_brfactor = 0;
		for (unsigned int x = 0; x < ntrans; x++) {
			fgets(stmp1, IOBUFFERSIZE, fp);
			sp1 = stmp1+1;
			sp2 = sp1;
			while ((*sp2) != ' ' && (*sp2) != ',') {
				sp2++;
			}
			memcpy (stmp2, sp1, sp2-sp1);
			stmp2[sp2-sp1] = '\0';
			src_state = (unsigned int) atoi(stmp2);
            if (j != src_state) {
				// new src state. add offset
				for (k = j; k < src_state; k++) {
					offsets[offsets_cnt++] = trans_cnt;
				}
				j = src_state;
				// update max. branching factor
				max_brfactor = MAX(max_brfactor, (trans_cnt - offsets[offsets_cnt-2]));
			}
			// read action label
			sp1 = sp2;
			while ((*sp1) == ' ' || (*sp1) == ',') {
				sp1++;
			}
			sp2 = strrchr(stmp1, ',');
			while ((*sp2) == ' ' || (*sp2) == ',') {
				sp2--;
			}
			sp2++;
			memcpy (stmp2, sp1, sp2-sp1);
			stmp2[sp2-sp1] = '\0';
			strcpy (act, stmp2);
			if (strncmp(act,"i",LABELSIZE) == 0 || strncmp(act,"\"tau\"",LABELSIZE) == 0 || strncmp(act,"tau",LABELSIZE) == 0 || strncmp(act,"\"i\"",LABELSIZE) == 0) {
				actid = 0;
			}
			else {
				actid = addact(act);
			}
			transact[trans_cnt] = actid;
			// read target state
			sp1 = sp2;
			while ((*sp1) == ' ' || (*sp1) == ',') {
				sp1++;
			}
			sp2 = sp1;
			while ((*sp2) != ' ' && (*sp2) != ')') {
				sp2++;
			}
			memcpy (stmp2, sp1, sp2-sp1);
			stmp2[sp2-sp1] = '\0';
			tgt_state = (unsigned int) atoi(stmp2);
			transtgt[trans_cnt++] = tgt_state;
		}
		// store final offset to end of trans array
        while (offsets_cnt < nstates+1) {
            offsets[offsets_cnt++] = trans_cnt;
		}
        // update max. branching factor
		max_brfactor = MAX(max_brfactor, (trans_cnt - offsets[offsets_cnt-1]));		
		fclose(fp);
		fprintf(stdout, "max. branching factor: %d\n", max_brfactor);

		// sort the transitions by action label. we use a permutation array for this.
		for (j = 0; j < nstates; j++) {
			size = offsets[j+1] - offsets[j];
			compare_offset = offsets[j];
			permutation = (unsigned int*) malloc(sizeof(unsigned int)*size);
			tmparray1 = (unsigned int*) malloc(sizeof(unsigned int)*size);
			tmparray2 = (unsigned int*) malloc(sizeof(unsigned int)*size);
			for (k = 0; k < size; k++) {
				permutation[k] = k;
			}
			// do the sorting
			qsort (permutation, size, sizeof(unsigned int), compare);
			// reorder based on final permutation
			for (k = 0; k < size; k++) {
				tmparray1[k] = transact[offsets[j]+permutation[k]];
				tmparray2[k] = transtgt[offsets[j]+permutation[k]];
			}
			for (k = 0; k < size; k++) {
				transact[offsets[j]+k] = tmparray1[k];
				transtgt[offsets[j]+k] = tmparray2[k];
			}
			free(permutation);
			free(tmparray1);
			free(tmparray2);
		}
		fprintf (stdout, "writing new .AUT file\n");
		strcpy(fn2, fn);
		sp1 = strrchr(fn2, '.');
		(*sp1) = '\0';
		strcat(fn2, "_n.aut");
		fp = fopen (fn2, "w");
		fprintf (fp, "des (0, %d, %d)\n", ntrans, nstates);
		for (j = 0; j < nstates; j++) {
			for (k = offsets[j]; k < offsets[j+1]; k++) {
				if (transact[k] == 0) {
					fprintf (fp, "(%d, i, %d)\n", j, transtgt[k]);
				}
				else {
					fprintf (fp, "(%d, %d, %d)\n", j, transact[k], transtgt[k]);						
				}
			}
		}
		fclose(fp);
		fprintf (stdout, "finished!\n");
		free(transact);
		free(transtgt);
		free(offsets);
		clean_actset();
	}
	else {
		fprintf(stderr, "ERROR: input file does not exist!\n");
		exit(1);
	}
}
