/* *
 * Copyright 1993-2014 NVIDIA Corporation.  All rights reserved.
 *
 * Please refer to the NVIDIA end user license agreement (EULA) associated
 * with this source code for terms and conditions that govern your use of
 * this software. Any use, reproduction, disclosure, or distribution of
 * this software and related documentation outside the terms of the EULA
 * is strictly prohibited.
 */

/* *
 * GPU Strong Bisimulation Reduction, based on the work of Rajasekaran & Lee,'94
 * / journal paper from '98 (Kanellakis/Smolka based). Written by Anton Wijs (2014).
 */

#include <iostream>
#include <fstream>
#include <sstream>
#include <vector>
#include <stdlib.h>
#include <cuda.h>
#include <cuda_runtime_api.h>
#include <assert.h>
#include <time.h>
#include <math.h>
#include <thrust/host_vector.h>
#include <thrust/device_vector.h>
#include <thrust/scan.h>
#include <thrust/sequence.h>
#include <thrust/copy.h>
#include <thrust/gather.h>
#include <thrust/sort.h>
#include "uthash.h"

using namespace std;

static const int BLOCK_SIZE = 512;
static const uint32_t EMPTYTRANS = 0x7FFFFFFF;
//static const uint32_t LOCKED_EMPTYSTATE = 0xFFFFFFFF;
static const int WARP_SIZE = 32;
const size_t Mb = 1<<20;
// BIT MANIPULATION MACROS

#define MIN(a,b) \
   ({ __typeof__ (a) _a = (a); \
       __typeof__ (b) _b = (b); \
     _a < _b ? _a : _b; })

#define MAX(a,b) \
   ({ __typeof__ (a) _a = (a); \
       __typeof__ (b) _b = (b); \
     _a > _b ? _a : _b; })

// BITS USED FOR BOOKKEEPING:
// 1 PER D_RESULTS ENTRY FOR REMOVING A STATE FROM THE COMPUTATION (IS PART OF STABLE BLOCK)
// 1 PER TAU D_TRANS ENTRY TO INDICATE THAT TAU-TRANSITION IS NOT CONFLUENT

#define PRINTTHREAD(j, i)						{printf("%d: Seen by thread %d: %d\n", (j), (blockIdx.x*blockDim.x)+threadIdx.x, (i));}

#define SETBITS(i, j, x)						for (bi = (i); bi < (j); bi++) { (x) = (x) | (((uint32_t) 1)<<bi); }
// --- BIT OPERATIONS FOR LOCKING
#define LOCK(t)									{(t) = (t) | 0x80000000;}
#define ISLOCKED(a, t)							{(a) = (((t) >> 31) > 0) ? 1 : 0;}
#define UNLOCK(t)								{(t) = (t) & 0x7FFFFFFF;}
#define LOCKED(t)								((t) | 0x80000000)
#define STRIPLOCK(t)							((t) & 0x7FFFFFFF)

#define THREADID								threadIdx.x
#define GLOBAL_THREADID							(blockIdx.x * blockDim.x) + threadIdx.x
#define GLOBAL_WARPID							(blockIdx.x * (blockDim.x / WARP_SIZE)) + (threadIdx.x / WARP_SIZE)
#define WARPID									threadIdx.x / WARP_SIZE
#define WARPTID									threadIdx.x % WARP_SIZE
#define WARPLEADERID							((threadIdx.x / WARP_SIZE) * WARP_SIZE)
#define GLOBAL_LASTINWARP_TID					GLOBAL_THREADID + (31 - WARPTID)
#define NROFTHREADS								gridDim.x * blockDim.x
#define SEARCH_NROFTHREADS						((NROFTHREADS) / 2)
#define SEARCH_THREADID							(GLOBAL_THREADID) % (SEARCH_NROFTHREADS)
//#define SEARCH_NROFTHREADS						(gridDim.x * blockDim.x) / 2

// macros for broadcasting in warp
#define BROADCAST_IF_WARPED						{	warpcont = thcont; \
													if (STEP_SIZE == WARP_SIZE) { \
														shared[THREADID] = thcont; \
														if (warpcont == 0) { \
															for (wj = WARPLEADERID; wj < WARPLEADERID + WARP_SIZE; wj++) { \
																warpcont = shared[wj]; \
																if (warpcont == 1) { \
																	break; \
																} \
															} \
														} \
													} \
												}
#define BROADCAST								{	warpcont = thcont; \
													shared[THREADID] = thcont; \
													if (warpcont == 0) { \
														for (wj = WARPLEADERID; wj < WARPLEADERID + WARP_SIZE; wj++) { \
															warpcont = shared[wj]; \
															if (warpcont == 1) { \
																break; \
															} \
														} \
													} \
												}

// definitions for signature comparisons
#define STEP_SIZE						((i < d_states_endofsegments) ? WARP_SIZE : 1)
#define PIVOT_STEP_SIZE					((candidates_buffer[ci] < d_states_endofsegments) ? WARP_SIZE : 1)
#define NR_OF_I_BUFFERS					(((trans_end - trans_start)/STEP_SIZE) + (THREAD_BUFFER_SIZE-1)) / THREAD_BUFFER_SIZE
#define NR_OF_PIVOT_BUFFERS				(((pivot_trans_end - pivot_trans_start)/PIVOT_STEP_SIZE) + (THREAD_BUFFER_SIZE-1)) / THREAD_BUFFER_SIZE
#define NOT_RELEVANT					2
#define STILL_POSSIBLE					3
#define TAU								0
// macro to enable/disable smart pivot candidate selection depending on label of first outgoing transition
#define LABELDEPENDENCY					first_label
//#define LABELDEPENDENCY					0

// macros for broadcasting in warp; shfl version (CUDA 3)
/* #define BROADCAST_IF_WARPED						{	warpcont = thcont; \
													if (STEP_SIZE == WARP_SIZE) { \
														for (j = 0; j < WARP_SIZE; j++) { \
															warpcont |= (uint32_t) __shfl((int) thcont, (int) j); \
														} \
													} \
												}
#define BROADCAST								{	warpcont = thcont; \
													for (j = 0; j < WARP_SIZE; j++) { \
														warpcont |= (uint32_t) __shfl((int) thcont, (int) j); \
													} \
												} */

// start indices of the different parts of the data
#define s_cont				0
#define s_block_id			1
#define s_offsets			2
#define s_trans_labels		nstates+3
#define s_trans_tgts		nstates+3+ntrans

/**
 * This macro checks return value of the CUDA runtime call and exits
 * the application if the call failed.
 */
#define CUDA_CHECK_RETURN(value) {											\
	cudaError_t _m_cudaStat = value;										\
	if (_m_cudaStat != cudaSuccess) {										\
		fprintf(stderr, "Error %s at line %d in file %s\n",					\
				cudaGetErrorString(_m_cudaStat), __LINE__, __FILE__);		\
		exit(1);															\
	} }

// TEST FUNCTIONS

int vmem = 0;

// GPU CONSTANTS
__constant__ uint32_t d_nstates;
__constant__ uint32_t d_ntrans;

//wrapper around cudaMalloc to count allocated memory and check for error while allocating
int cudaMallocCount ( void ** ptr,int size) {
	cudaError_t err = cudaSuccess;
	vmem += size;
	err = cudaMalloc(ptr,size);
	if (err) {
		printf("Error %s at line %d in file %s\n", cudaGetErrorString(err), __LINE__, __FILE__);
		exit(1);
	}
	fprintf (stdout, "allocated %d\n", size);
	return size;
}

//// test function, printing the content of the global memory on the screen
//void print_status(uint32_t *d_offsets, uint32_t *d_trans_labels, uint32_t *d_trans_tgts, uint32_t *d_results, uint32_t nstates, uint32_t ntrans) {
//	uint32_t *tst_offsets = (uint32_t *) malloc((nstates+1) * sizeof(uint32_t));
//	uint32_t *tst_trans_labels = (uint32_t *) malloc(ntrans * sizeof(uint32_t));
//	uint32_t *tst_trans_tgts = (uint32_t *) malloc(ntrans * sizeof(uint32_t));
//	uint32_t *tst_results = (uint32_t *) malloc(nstates * sizeof(uint32_t));
//	uint32_t result;
//
//	CUDA_CHECK_RETURN(cudaMemcpy(tst_offsets, d_offsets, (nstates+1) * sizeof(uint32_t), cudaMemcpyDeviceToHost));
//	CUDA_CHECK_RETURN(cudaMemcpy(tst_trans_labels, d_trans_labels, ntrans * sizeof(uint32_t), cudaMemcpyDeviceToHost));
//	CUDA_CHECK_RETURN(cudaMemcpy(tst_trans_tgts, d_trans_tgts, ntrans * sizeof(uint32_t), cudaMemcpyDeviceToHost));
//	CUDA_CHECK_RETURN(cudaMemcpy(tst_results, d_results, nstates * sizeof(uint32_t), cudaMemcpyDeviceToHost));
//	
//	fprintf (stdout, "offsets (with lock info):\n");
//	for (int i = 0; i < nstates+1; i++) {
//		ISLOCKED(result, tst_offsets[i]);
//		fprintf(stdout, "(%d,%d)  ", result, STRIPLOCK(tst_offsets[i]));
//		fprintf (stdout, "\n");
//	}
//	fprintf (stdout, "trans labels (with lock info):\n");
//	for (int i = 0; i < ntrans; i++) {
//		ISLOCKED(result, tst_trans_labels[i]);
//		fprintf(stdout, "(%d,%d)  ", result, STRIPLOCK(tst_trans_labels[i]));
//		fprintf (stdout, "\n");
//	}
//	fprintf (stdout, "trans targets (with lock info):\n");
//	for (int i = 0; i < ntrans; i++) {
//		ISLOCKED(result, tst_trans_tgts[i]);
//		fprintf(stdout, "(%d,%d)  ", result, STRIPLOCK(tst_trans_tgts[i]));
//		fprintf (stdout, "\n");
//	}
//	fprintf (stdout, "results (with lock info):\n");
//	for (int i = 0; i < nstates; i++) {
//		ISLOCKED(result, tst_results[i]);
//		fprintf(stdout, "(%d,%d)  ", result, STRIPLOCK(tst_results[i]));
//		fprintf (stdout, "\n");
//	}
//}

// hashable struct to count the number of states in the reduced LTS
typedef struct statenode {
	int id;
	UT_hash_handle hh;
} statenode;

// hashtable
statenode *stateset;
uint32_t statesetcount = 0;

// clean the hashtable
void clean_stateset() {
	statenode *current, *tmp;
	
	HASH_ITER(hh, stateset, current, tmp) {
		HASH_DEL(stateset, current);
		free(current);
	}
	statesetcount = 0;
}

// print the hashtable content
void print_stateset() {
	statenode *current, *tmp;
	
	HASH_ITER(hh, stateset, current, tmp) {
		fprintf (stdout, "%d ", current->id);
	}
	fprintf (stdout, "\n");
}

// add a state to the stateset;
void addstate(uint32_t id) {
	statenode *s;
	int statecast = (int) id;
	HASH_FIND_INT(stateset, &statecast, s);
	if (s==NULL) {
		s = (statenode*) malloc(sizeof(statenode));
		s->id = statecast;
		HASH_ADD_INT(stateset, id, s);
		statesetcount++;
	}
}

// test function, printing the number of states in the output LTS on the screen
void print_outputsize(uint32_t *d_results, uint32_t nstates) {
	uint32_t *tst_results = (uint32_t *) malloc(nstates * sizeof(uint32_t));
	CUDA_CHECK_RETURN(cudaMemcpy(tst_results, d_results, nstates * sizeof(uint32_t), cudaMemcpyDeviceToHost));

	clean_stateset();
	for (int i = 0; i < nstates; i++) {
		addstate(tst_results[i]);
	}
	fprintf (stdout, "nr of states in reduced LTS: %d\n", statesetcount);
	print_stateset();
	//print_blocks(tst_results, nstates);
	clean_stateset();
}

// test function, printing the number of states in the output LTS on the screen
void print_reprs(uint32_t *d_V_array, uint32_t *d_partition_s_array, uint32_t *d_temp_array, uint32_t *d_L_src_array, uint32_t *d_L_lbl_array, uint32_t *d_L_tgt_array, uint32_t nstates, uint32_t ntrans) {
	uint32_t *V_array = (uint32_t *) malloc(nstates * sizeof(uint32_t));
	uint32_t *partition_s_array = (uint32_t *) malloc(nstates * sizeof(uint32_t));
	uint32_t *temp_array = (uint32_t *) malloc((nstates+1) * sizeof(uint32_t));
	uint32_t *L_src_array = (uint32_t *) malloc(ntrans * sizeof(uint32_t));
	uint32_t *L_lbl_array = (uint32_t *) malloc(ntrans * sizeof(uint32_t));
	uint32_t *L_tgt_array = (uint32_t *) malloc(ntrans * sizeof(uint32_t));
	CUDA_CHECK_RETURN(cudaMemcpy(V_array, d_V_array, nstates * sizeof(uint32_t), cudaMemcpyDeviceToHost));
	CUDA_CHECK_RETURN(cudaMemcpy(partition_s_array, d_partition_s_array, nstates * sizeof(uint32_t), cudaMemcpyDeviceToHost));
	CUDA_CHECK_RETURN(cudaMemcpy(temp_array, d_temp_array, (nstates+1) * sizeof(uint32_t), cudaMemcpyDeviceToHost));
	CUDA_CHECK_RETURN(cudaMemcpy(L_src_array, d_L_src_array, ntrans * sizeof(uint32_t), cudaMemcpyDeviceToHost));
	CUDA_CHECK_RETURN(cudaMemcpy(L_lbl_array, d_L_lbl_array, ntrans * sizeof(uint32_t), cudaMemcpyDeviceToHost));
	CUDA_CHECK_RETURN(cudaMemcpy(L_tgt_array, d_L_tgt_array, ntrans * sizeof(uint32_t), cudaMemcpyDeviceToHost));

	uint32_t j;

	for (uint32_t i = 0; i < nstates; i++) {
		if (V_array[i] == 0) {
			fprintf(stdout, "%d: (%d,%d)\n", partition_s_array[i], temp_array[i], temp_array[i+1]);
			for (j = temp_array[i]; j < temp_array[i+1]; j++) {
				//ISLOCKED(result, L_src_array[j]);
				//if (result == 0) {
					fprintf (stdout, "%d-%d->%d ", L_src_array[j], L_lbl_array[j], L_tgt_array[j]);
				//}
			}
			fprintf (stdout, "\n");
		}
	}
}


// test function, print an array located in the GPU memory
void print_array(uint32_t *d_array, uint32_t size) {
	uint32_t *array = (uint32_t *) malloc(size * sizeof(uint32_t));
	CUDA_CHECK_RETURN(cudaMemcpy(array, d_array, size * sizeof(uint32_t), cudaMemcpyDeviceToHost));

	for (int i = 0; i < size; i++) {
		fprintf (stdout, "(%d,%d) ", i, array[i]);
	}
	fprintf (stdout, "\n");
}


__global__ void	fill_temp(uint32_t *d_offsets, uint32_t *d_partition_s, uint32_t *d_temp) {
	uint32_t state, offset1, offset2;
	
	for (int i = GLOBAL_THREADID; i < d_nstates; i += NROFTHREADS) {
		state = d_partition_s[i];
		// lookup offsets
		offset1 = d_offsets[state];
		offset2 = d_offsets[state+1];
		d_temp[i] = offset2 - offset1;
	}
}

//__global__ void initialise_V(uint32_t *d_partition_i_array, uint32_t *d_V_array) {	
//	if (GLOBAL_THREADID == 0) {
//		d_V_array[0] = 0;
//	}
//	for (int i = GLOBAL_THREADID+1; i < d_nstates; i += NROFTHREADS) {
//		if (d_partition_i_array[i] != d_partition_i_array[i-1]) {
//			d_V_array[i] = 0;
//		}
//		else {
//			d_V_array[i] = 1;
//		}
//	}
//}

__global__ void	store_transitions(uint32_t *d_offsets, uint32_t *d_trans_labels, uint32_t *d_trans_tgts, uint32_t *d_partition_s_array, uint32_t *d_temp_array,
					uint32_t *d_B_array, uint32_t *d_V_array, uint32_t *d_L_src_array, uint32_t *d_L_lbl_array, uint32_t *d_L_idx_array, uint32_t *d_L_tgt_array) {
	uint32_t state, offset1, offset2, j, k, Lstart, tgt, Vs, block;
	
	for (int i = GLOBAL_THREADID; i < d_nstates; i += NROFTHREADS) {
		state = d_partition_s_array[i];
		Lstart = d_temp_array[i];
		Vs = d_V_array[i];
		offset1 = d_offsets[state];
		offset2 = d_offsets[state+1];
		block = d_B_array[state];

		k = 0;
		for (j = offset1; j < offset2; j++) {
			d_L_src_array[Lstart+k] = block;
			d_L_lbl_array[Lstart+k] = d_trans_labels[j];
			d_L_idx_array[Lstart+k] = Vs;
			tgt = d_trans_tgts[j];
			d_L_tgt_array[Lstart+k] = d_B_array[tgt];
			k++;
		}
	}
}

__global__ void mark_duplicates(uint32_t *d_L_src_array, uint32_t *d_L_lbl_array, uint32_t *d_L_idx_array, uint32_t *d_L_tgt_array) {
	extern volatile __shared__ uint32_t shared[];

	uint32_t entry1, entry2, warpcont, wj;
	uint32_t thcont = 0;
	
	for (int i = GLOBAL_THREADID; i < (d_ntrans-1); i += NROFTHREADS) {
		entry1 = STRIPLOCK(d_L_src_array[i]);
		entry2 = STRIPLOCK(d_L_src_array[i+1]);
		thcont = (entry1 == entry2) ? 1 : 0;
		BROADCAST;
		if (warpcont == 1) {
			entry1 = d_L_idx_array[i];
			entry2 = d_L_idx_array[i+1];
			thcont = (thcont == 1 && entry1 == entry2) ? 1 : 0;
			BROADCAST;
			if (warpcont == 1) {
				entry1 = d_L_lbl_array[i];
				entry2 = d_L_lbl_array[i+1];
				thcont = (thcont == 1 && entry1 == entry2) ? 1 : 0;
				BROADCAST;
				if (warpcont == 1) {
					entry1 = d_L_tgt_array[i];
					entry2 = d_L_tgt_array[i+1];
					if (thcont == 1 && entry1 == entry2) {
						// entries are identical. Clean the neighbour by locking d_L_src_array[i+1]
						LOCK(d_L_src_array[i+1]);
					}
				}
			}
		}
	}
}

__global__ void compare_and_mark(uint32_t *d_partition_s_array, uint32_t *d_V_array, uint32_t *d_temp_array, uint32_t *d_B_array, uint32_t *d_L_src_array,
								uint32_t *d_L_lbl_array, uint32_t *d_L_idx_array, uint32_t *d_L_tgt_array, uint32_t *d_permutations_array, uint32_t *d_block_id, uint32_t *d_cont) {
	uint32_t state, j, pivot, offset1, offset2, pivot_offset1, pivot_offset2, entry1, entry2, result;
	
	for (int i = GLOBAL_THREADID; i < d_nstates; i += NROFTHREADS) {
		// obtain state id
		state = d_partition_s_array[i];
		// obtain index in block
		j = d_V_array[i];
		// compare signatures
		if (j > 0) {
			// pivot id to be found j positions earlier
			pivot = d_partition_s_array[i-j];
			// obtain offsets
			offset1 = d_temp_array[i];
			offset2 = d_temp_array[i+1];
			pivot_offset1 = d_temp_array[i-j];
			pivot_offset2 = d_temp_array[i-j+1];
			//if (state == 2524) {
			//	PRINTTHREAD(state, offset1);
			//	PRINTTHREAD(state, offset2);
			//	PRINTTHREAD(pivot, pivot_offset1);
			//	PRINTTHREAD(pivot, pivot_offset2);
			//}
			entry1 = EMPTYTRANS; entry2 = EMPTYTRANS;
			while (offset1 < offset2 && pivot_offset1 < pivot_offset2) {
				while (entry1 == EMPTYTRANS && offset1 < offset2) {
					entry1 = d_L_src_array[offset1];
					ISLOCKED(result, entry1);
					if (result == 1) {
						entry1 = EMPTYTRANS;
						offset1++;
					}
					UNLOCK(entry1);
				}
				while (entry2 == EMPTYTRANS && pivot_offset1 < pivot_offset2) {
					entry2 = d_L_src_array[pivot_offset1];
					ISLOCKED(result, entry2);
					if (result == 1) {
						entry2 = EMPTYTRANS;
						pivot_offset1++;
					}
					UNLOCK(entry2);
				}
				if (entry1 != EMPTYTRANS && entry2 != EMPTYTRANS) {
					//if (state == 2524) {
					//	PRINTTHREAD(0, entry1);
					//	PRINTTHREAD(1, entry2);
					//}
					// we have transitions to compare
//					if (entry1 == entry2) {
//						entry1 = d_L_idx_array[offset1];
//						entry2 = d_L_idx_array[pivot_offset1];
//						if (entry1 == entry2) {
					entry1 = d_L_lbl_array[offset1];
					entry2 = d_L_lbl_array[pivot_offset1];							
					if (entry1 == entry2) {
						//if (state == 2524) {
						//	PRINTTHREAD(2, entry1);
						//	PRINTTHREAD(3, entry2);
						//}
						entry1 = d_L_tgt_array[offset1];
						entry2 = d_L_tgt_array[pivot_offset1];
						if (entry1 == entry2) {
							//if (state == 2524) {
							//	PRINTTHREAD(4, entry1);
							//	PRINTTHREAD(5, entry2);
							//}
							offset1++; pivot_offset1++;
							entry1 = EMPTYTRANS;
							entry2 = EMPTYTRANS;
						}
						else {
							break;
						}													
					}
					else {
						break;
					}
					//	}
					//	else {
					//		break;
					//	}
					//}
					//else {
					//	break;
					//}
				}
			}
			// try to reach end of offsets, if not yet achieved
			while (offset1 < offset2) {
				entry1 = d_L_src_array[offset1];
				ISLOCKED(result, entry1);
				if (result == 0) {
					break;
				}
				offset1++;
			}
			while (pivot_offset1 < pivot_offset2) {
				entry1 = d_L_src_array[pivot_offset1];
				ISLOCKED(result, entry1);
				if (result == 0) {
					break;
				}
				pivot_offset1++;
			}
			// if offsets are not equal, there is a problem
			if (offset1 != offset2 || pivot_offset1 != pivot_offset2) {
				//if (state == 3) {
				//	PRINTTHREAD(offset1, offset2);
				//	PRINTTHREAD(pivot_offset1, pivot_offset2);
				//}
				//// add the state entry to a new block
				////entry1 = (*d_block_id);
				////d_B_array[state] = entry1;
				////d_partition_i_array[state] = entry1;
				////(*d_cont) = 1;
				
				// mark state entry in B
				entry1 = d_B_array[state];
				d_B_array[state] = LOCKED(entry1);
				// mark entry corresponding with pivot in permutations atomically. If successful, increase block_id counter and store value in permutations
				entry1 = d_permutations_array[pivot];
				ISLOCKED(entry2, entry1);
				if (entry2 == 0) {
					// try to claim the entry
					entry2 = atomicCAS((uint32_t *) d_permutations_array+pivot, entry1, LOCKED(entry1));
					if (entry2 == entry1) {
						// successful. increment block_id to obtain new id.
						entry2 = atomicAdd((uint32_t *) d_block_id, 1);
						// store new id
						entry2++;
						d_permutations_array[pivot] = LOCKED(entry2);
						// indicate globally that a next iteration of algorithm is required
						(*d_cont) = 1;
					}
				}
			}
		}
	}
}

__global__ void update_block_info(uint32_t *d_partition_s_array, uint32_t *d_partition_i_array, uint32_t *d_V_array, uint32_t *d_permutations_array, uint32_t *d_B_array) {
	uint32_t state, j, pivot, bid, result;
	
	for (int i = GLOBAL_THREADID; i < d_nstates; i += NROFTHREADS) {
		state = d_partition_s_array[i];
		j = d_V_array[i];
		bid = d_B_array[state];

		ISLOCKED(result, bid);
		if (result == 1) {
			// entry locked, so this state will be added to new block. ID to be found in d_permutations.
			pivot = d_partition_s_array[i-j];
			bid = d_permutations_array[pivot];
			UNLOCK(bid);
			// update block id info
			d_B_array[state] = bid;
			d_partition_i_array[i] = bid;
		}
	}
}

void update_permutation(thrust::device_vector<uint32_t>& keys, thrust::device_vector<uint32_t>& permutation) {
    // temporary storage for keys
    thrust::device_vector<uint32_t> temp(keys.size());

    // permute the keys with the current reordering
    thrust::gather(permutation.begin(), permutation.end(), keys.begin(), temp.begin());

    // stable_sort the permuted keys and update the permutation
    thrust::stable_sort_by_key(temp.begin(), temp.end(), permutation.begin());
}

void apply_permutation(thrust::device_vector<uint32_t>& keys, thrust::device_vector<uint32_t>& permutation) {
    // copy keys to temporary vector
    thrust::device_vector<uint32_t> temp(keys.begin(), keys.end());

    // permute the keys
    thrust::gather(permutation.begin(), permutation.end(), temp.begin(), keys.begin());
}

/**
 * Host function that reads AUT file (LTS), prepares data for GPU, and launches reduction procedure.
 */
int main(int argc, char** argv) {
	int nblocks = 1;
	int nblocksfortrans = 1;
	int nthreadsperblock = BLOCK_SIZE;
	// level of verbosity (1=print progress)
	int verbosity = 0;
	// clock to measure time
	clock_t start, stop;
	double runtime = 0.0;
	string fn;
	uint32_t i, j, k;

	// Structures to store LTS input
	uint32_t *offsets, *trans_labels, *trans_tgts;
	uint32_t nstates, ntrans;
	uint32_t cont;

	cudaDeviceProp prop;
	int nDevices;

	// GPU side versions of the input
	uint32_t *d_data;

	if (argc == 1) {
		fprintf(stderr, "ERROR: No input AUT file given!\n");
		exit(1);
	}
	fn = string(argv[1]);

	i = 2;
	while (i < argc) {
		if (!strcmp(argv[i],"-b")) {
			// store nr of blocks to be used
			nblocks = atoi(argv[i+1]);
			i += 2;
		}
		else if (!strcmp(argv[i],"-t")) {
			// store nr of threads per block to be used
			nthreadsperblock = atoi(argv[i+1]);
			i += 2;
		}
		else if (!strcmp(argv[i],"-v")) {
			// store verbosity level
			verbosity = atoi(argv[i+1]);
			if (verbosity > 3) {
				verbosity = 3;
			}
			i += 2;
		}
	}

	ifstream fp (fn.c_str()); 

	// read file header
	string s;
	getline(fp, s);
	// strip 'des(', ')'
	s = s.substr(4).substr(0,s.size()-5);

	istringstream ss(s);
	vector <string> data;
	
	while (ss) {
		string s;
		if (!getline (ss, s, ',')) break;
		if (s[0] == ' ') {
			s = s.substr(1);
		}
		data.push_back(s);
	}
	// process header
	nstates = (uint32_t) atoi(data[2].c_str());
	fprintf(stdout, "nr of states in LTS: %d\n", nstates);
	ntrans = (uint32_t) atoi(data[1].c_str());
	fprintf(stdout, "nr of transitions in LTS: %d\n", ntrans);

	// current offset index
	i = 0;
	// current trans index
	j = 0;
	// last source
	k = 0;
	// fill the arrays with the LTS transitions
	offsets = (uint32_t*) malloc(sizeof(uint32_t)*(nstates+1));
	trans_labels = (uint32_t*) malloc(sizeof(uint32_t)*ntrans);
	trans_tgts = (uint32_t*) malloc(sizeof(uint32_t)*ntrans);
	offsets[i++] = 0;

	while (fp) {
		// Read the input
		string s;
		if (!getline(fp, s)) break;
		
		// strip '(', ')'
		s = s.substr(1).substr(0,s.size()-2);

		istringstream ss(s);
		vector <string> data;
		
		while (ss) {
			string s;
			if (!getline (ss, s, ',')) break;
			if (s[0] == ' ') {
				s = s.substr(1);
			}
			data.push_back(s);
		}
		while (atoi(data[0].c_str()) > k) {
			offsets[i++] = j;
			k++;
		}
		// store label id
		// is label tau?
		if (data[1] == "i") {
			trans_labels[j] = 0;
		}
		else {
			// strip '"', '"'
			//data[1] = data[1].substr(1).substr(0,data[1].size()-2);
			trans_labels[j] = atoi(data[1].c_str());
		}
		// store target
		trans_tgts[j++] = atoi(data[2].c_str());
	}
	// finalise offsets (take care of deadlock states at the end)
	while (i < nstates+1) {
		offsets[i++] = j;
	}

	// Query the device properties and determine max number of integers that we can work with
	cudaGetDeviceCount(&nDevices);
	if (nDevices == 0) {
		fprintf (stderr, "ERROR: No CUDA compatible GPU detected!\n");
		exit(1);
	}
	cudaGetDeviceProperties(&prop, 0);
	fprintf (stdout, "global mem: %lu\n", (size_t) prop.totalGlobalMem);
	fprintf (stdout, "shared mem per block: %d\n", (int) prop.sharedMemPerBlock);
	fprintf (stdout, "max. threads per block: %d\n", (int) prop.maxThreadsPerBlock);
	fprintf (stdout, "max. grid size: %lu\n", (size_t) prop.maxGridSize[0]);
	fprintf (stdout, "nr. of multiprocessors: %d\n", (int) prop.multiProcessorCount);

	size_t available, total;
	cudaMemGetInfo(&available, &total);
	
	size_t q_size = 0;
	if (q_size == 0) {
		q_size = total / sizeof(uint32_t);
	}
	size_t el_per_Mb = Mb / sizeof(uint32_t);
	
	while(cudaMalloc((void**)&d_data,  q_size * sizeof(uint32_t)) == cudaErrorMemoryAllocation)	{			
		q_size -= el_per_Mb;
		if( q_size  < el_per_Mb) {
			// signal no free memory
			break;
		}
	}
	// now that we know q_size, free memory again
	cudaFree(d_data);
	fprintf (stdout, "max. nr. of integers that can be stored on GPU: %lu\n", q_size);
	// compute amount of memory required for input
	j = nstates+2*ntrans+3;
	if (j > q_size) {
		fprintf (stderr, "not enough memory on GPU available for input!\n");
		exit(-1);
	}
	// determine actual nr of blocks
	if (nblocks == 1) {
		nblocks = MAX(1, (nstates + (nthreadsperblock - 1))/ nthreadsperblock);
	}
	if (nblocks > 0) {
		nblocks = MAX(1,MIN(prop.maxGridSize[0],nblocks));
	}
	else {
		// set to maximum nr
		nblocks = prop.maxGridSize[0];
	}
	fprintf (stdout, "%d blocks will be launched on the GPU\n", nblocks);
	// determine actual nr of threads per block
	nthreadsperblock = MAX(1,MIN(prop.maxThreadsPerBlock,nthreadsperblock));
	// determine nr of blocks when parallelising on the transitions
	nblocksfortrans = MAX(1,MIN(prop.maxGridSize[0], ntrans / nthreadsperblock));

	fprintf (stdout, "launching reduction\n");
	// Start timer
	assert((start = clock())!=-1);
	// Allocate memory on GPU
	cudaMalloc((void**)&d_data,  j * sizeof(uint32_t));

	// GPU Structures for algorithm
	thrust::device_vector<uint32_t> d_temp(nstates+1);
	thrust::device_vector<uint32_t> d_L_src(ntrans);
	thrust::device_vector<uint32_t> d_L_lbl(ntrans);
	thrust::device_vector<uint32_t> d_L_idx(ntrans);
	thrust::device_vector<uint32_t> d_L_tgt(ntrans);
	// vector needed to sort the transition vectors
	thrust::device_vector<uint32_t> d_permutations(ntrans);
	// contains info on block membership per state. Initially, we have one large block with ID 0
	thrust::device_vector<uint32_t> d_B(nstates, 0);
	thrust::device_vector<uint32_t> d_partition_i(nstates, 0);
	thrust::device_vector<uint32_t> d_partition_s(nstates);
	// vector containing indices per state within their blocks. Not mentioned by Rajasekaran & Lee in their
	// pseudo-code to be computed in each iteration!!!
	thrust::device_vector<uint32_t> d_V(nstates);
	
	// pointers to thrust vectors for kernels
	uint32_t* d_temp_array = thrust::raw_pointer_cast( &d_temp[0] );
	uint32_t* d_partition_i_array = thrust::raw_pointer_cast( &d_partition_i[0] );
	uint32_t* d_partition_s_array = thrust::raw_pointer_cast( &d_partition_s[0] );
	uint32_t* d_L_src_array = thrust::raw_pointer_cast( &d_L_src[0] );
	uint32_t* d_L_lbl_array = thrust::raw_pointer_cast( &d_L_lbl[0] );
	uint32_t* d_L_idx_array = thrust::raw_pointer_cast( &d_L_idx[0] );
	uint32_t* d_L_tgt_array = thrust::raw_pointer_cast( &d_L_tgt[0] );
	uint32_t* d_B_array = thrust::raw_pointer_cast( &d_B[0] );
	uint32_t* d_V_array = thrust::raw_pointer_cast( &d_V[0] );
	uint32_t* d_permutations_array = thrust::raw_pointer_cast( &d_permutations[0] );	
	
	// Copy data to GPU
	CUDA_CHECK_RETURN(cudaMemcpy(&(d_data[s_offsets]), offsets, (nstates+1)*sizeof(uint32_t), cudaMemcpyHostToDevice))
	CUDA_CHECK_RETURN(cudaMemcpy(&(d_data[s_trans_labels]), trans_labels, ntrans*sizeof(uint32_t), cudaMemcpyHostToDevice))
	CUDA_CHECK_RETURN(cudaMemcpy(&(d_data[s_trans_tgts]), trans_tgts, ntrans*sizeof(uint32_t), cudaMemcpyHostToDevice))
	// copy symbols
	cudaMemcpyToSymbol(d_nstates, &nstates, sizeof(uint32_t));
	cudaMemcpyToSymbol(d_ntrans, &ntrans, sizeof(uint32_t));

	uint32_t zero = 0 ;

	// start of Rajasekaran/Lee algorithm
	
	// initialise variables
	CUDA_CHECK_RETURN(cudaMemcpy(&(d_data[s_cont]), &zero, sizeof(uint32_t), cudaMemcpyHostToDevice));
	CUDA_CHECK_RETURN(cudaMemcpy(&(d_data[s_block_id]), &zero, sizeof(uint32_t), cudaMemcpyHostToDevice));

	// set the elements of d_partition_s to 0, 1, 2, 3, ...
	thrust::sequence(d_partition_s.begin(), d_partition_s.end());

	cont = 1;
	uint32_t counter = 0;
	while (cont == 1) {
		CUDA_CHECK_RETURN(cudaMemcpy(&(d_data[s_cont]), &zero, sizeof(uint32_t), cudaMemcpyHostToDevice));
		//fprintf (stdout, "1\n");
		// fill the temp vector
		fill_temp<<<nblocks, nthreadsperblock>>>(&(d_data[s_offsets]), d_partition_s_array, d_temp_array);
		//print_array(d_partition_s_array, nstates);
		//print_array(d_temp_array, nstates);
		//fprintf (stdout, "2\n");
		// perform exclusive scan
		thrust::exclusive_scan(d_temp.begin(), d_temp.begin()+nstates+1, d_temp.begin());
		//fprintf (stdout, "3\n");
		// compute V (index for each state within its block)
		thrust::fill(d_V.begin(), d_V.end(), 1);
		//fprintf (stdout, "4\n");
		//initialise_V<<<nblocks, nthreadsperblock>>>(d_partition_i_array, d_V_array);
		// perform exclusive scan on V to finish computing it
		thrust::exclusive_scan_by_key(d_partition_i.begin(), d_partition_i.begin() + nstates, d_V.begin(), d_V.begin());
		//print_array(d_partition_i_array, nstates);
		//print_array(d_V_array, nstates);
		//fprintf (stdout, "5\n");
		// store transitions info in the L vectors
		//print_array(d_temp_array, nstates+1);
		store_transitions<<<nblocks, nthreadsperblock>>>(&(d_data[s_offsets]), &(d_data[s_trans_labels]), &(d_data[s_trans_tgts]), d_partition_s_array, d_temp_array, d_B_array, d_V_array, d_L_src_array, d_L_lbl_array, d_L_idx_array, d_L_tgt_array);
		//print_array(d_V_array, nstates);
		//fprintf (stdout, "6\n");
		// sort the transitions in lexicographic order
		thrust::sequence(d_permutations.begin(), d_permutations.end());
		update_permutation(d_L_tgt, d_permutations);
		update_permutation(d_L_lbl, d_permutations);
		update_permutation(d_L_idx, d_permutations);
		update_permutation(d_L_src, d_permutations);
		//fprintf (stdout, "7\n");
		// apply final permutation
		apply_permutation(d_L_tgt, d_permutations);
		apply_permutation(d_L_lbl, d_permutations);
		apply_permutation(d_L_idx, d_permutations);
		apply_permutation(d_L_src, d_permutations);
		//print_array(d_L_src_array, ntrans);
		//print_array(d_L_lbl_array, ntrans);
		//print_array(d_L_idx_array, ntrans);
		//print_array(d_L_tgt_array, ntrans);
		//fprintf (stdout, "8\n");
		// mark duplicates in the transitions
		mark_duplicates<<<nblocksfortrans, nthreadsperblock, nthreadsperblock * sizeof(uint32_t)>>>(d_L_src_array, d_L_lbl_array, d_L_idx_array, d_L_tgt_array);
		//print_array(d_L_src_array, ntrans);
		//print_array(d_L_lbl_array, ntrans);
		//print_array(d_L_idx_array, ntrans);
		//print_array(d_L_tgt_array, ntrans);
		//fprintf (stdout, "9\n");
		//print_array(d_temp_array, nstates+1);
		//print_reprs(d_V_array, d_partition_s_array, d_temp_array, d_L_src_array, d_L_lbl_array, d_L_tgt_array, nstates, ntrans);
		// compare signatures
		compare_and_mark<<<nblocks, nthreadsperblock>>>(d_partition_s_array, d_V_array, d_temp_array, d_B_array, d_L_src_array, d_L_lbl_array, d_L_idx_array, d_L_tgt_array, d_permutations_array, &(d_data[s_block_id]), &(d_data[s_cont]));
		//fprintf(stdout, "offsets:\n");
		//print_array(d_temp_array, nstates+1);
		//print_array(d_partition_s_array, nstates);
		//print_array(d_B_array, nstates);
		//print_array(d_permutations_array, nstates);
		//fprintf (stdout, "10\n");
		// update block info
		update_block_info<<<nblocks, nthreadsperblock>>>(d_partition_s_array, d_partition_i_array, d_V_array, d_permutations_array, d_B_array);
		//print_array(d_B_array, nstates);
		//fprintf (stdout, "11\n");
		// sort the partition vectors to reflect the change
		thrust::stable_sort_by_key(d_partition_i.begin(), d_partition_i.end(), d_partition_s.begin());
		//fprintf(stdout, "result:\n");
		//print_array(d_partition_i_array, nstates);
		//print_array(d_partition_s_array, nstates);
		//print_array(d_B_array, nstates);		
		//fprintf (stdout, "12\n");

		CUDA_CHECK_RETURN(cudaMemcpy(&cont, &(d_data[s_cont]), sizeof(uint32_t), cudaMemcpyDeviceToHost));
		//if (cont == 1) {
		//	value++;
		//	CUDA_CHECK_RETURN(cudaMemcpy(&(d_data[s_block_id]), &value, sizeof(uint32_t), cudaMemcpyHostToDevice));
		//}
		//for(i = 0; i < d_temp.size(); i++) std::cout << "d_temp[" << i << "] = " << d_temp[i] << std::endl;
		//print_outputsize(d_B_array, nstates);
		//fprintf (stdout, "%d\n", counter++);
	}
	// determine runtime
	stop = clock();
	runtime = (double) (stop-start)/CLOCKS_PER_SEC;
	print_outputsize(d_B_array, nstates);
	fprintf (stdout, "Run time: %f\n", runtime);
	return 0;
}
