/* *
 * Copyright 1993-2014 NVIDIA Corporation.  All rights reserved.
 *
 * Please refer to the NVIDIA end user license agreement (EULA) associated
 * with this source code for terms and conditions that govern your use of
 * this software. Any use, reproduction, disclosure, or distribution of
 * this software and related documentation outside the terms of the EULA
 * is strictly prohibited.
 */

/* *
 * GPU Strong and Branching Bisimulation Reduction. Written by Anton Wijs (2014).
 */

#include <iostream>
#include <fstream>
#include <sstream>
#include <string>
#include <vector>
#include <stdlib.h>
#include <stdint.h>
#include <cuda.h>
#include <cuda_runtime_api.h>
#include <assert.h>
#include <time.h>
#include <math.h>
#include "uthash.h"

using namespace std;

static const int BLOCK_SIZE = 512;
static const uint32_t EMPTYTRANS = 0x7FFFFFFF;
static const uint32_t NOBLOCK = 0x3FFFFFFF;
static const uint32_t LOCKED_NOBLOCK = 0xBFFFFFFF;
//static const uint32_t LOCKED_EMPTYSTATE = 0xFFFFFFFF;
static const int THREAD_BUFFER_SIZE = 20;
static const int WARP_SIZE = 32;
const size_t Mb = 1<<20;
// BIT MANIPULATION MACROS

#define MIN(a,b) \
   ({ __typeof__ (a) _a = (a); \
       __typeof__ (b) _b = (b); \
     _a < _b ? _a : _b; })

#define MAX(a,b) \
   ({ __typeof__ (a) _a = (a); \
       __typeof__ (b) _b = (b); \
     _a > _b ? _a : _b; })

// BITS USED FOR BOOKKEEPING:
// 1 PER D_RESULTS ENTRY FOR REMOVING A STATE FROM THE COMPUTATION (IS PART OF STABLE BLOCK)
// 1 PER TAU D_TRANS ENTRY TO INDICATE THAT TAU-TRANSITION IS NOT CONFLUENT

#define PRINTTHREAD(j, i)						{printf("%d: Seen by thread %d: %d\n", (j), (blockIdx.x*blockDim.x)+threadIdx.x, (i));}

#define SETBITS(i, j, x)						for (bi = (i); bi < (j); bi++) { (x) = (x) | (((uint32_t) 1)<<bi); }
// --- BIT OPERATIONS TO BE PERFORMED ON D_RESULTS ENTRIES
#define GETREPRESENTATIVE(a, t)					{(a) = (t) & 0x7FFFFFFF;}
#define SETBLOCKPROPAGATED(t)					{(t) = (t) | 0x80000000;}
#define RESETBLOCKPROPAGATED(t)					{(t) = (t) & 0x7FFFFFFF;}
#define BLOCKISPROPAGATED(a, t)					{(a) = (((t) >> 31) > 0) ? 1 : 0;}
//#define SETPROP(t)								{(t) = (t) | 0x40000000;}
//#define REMOVEPROP(t)							{(t) = (t) & 0xBFFFFFFF;}
//#define ISPROPAGATED(a, t)						{(a) = (((t) & 0x40000000) > 0) ? 1 : 0;}
// --- BIT OPERATIONS TO BE PERFORMED ON D_OFFSETS ENTRIES
#define GETOFFSET(a, t)							{(a) = (t) & 0x7FFFFFFF;}
#define MAKEPIVOTCANDIDATE(t)					{(t) = (t) | 0x80000000;}
#define ISPIVOTCANDIDATE(a, t)					{(a) = (((t) >> 31) > 0) ? 1 : 0;}
#define REMOVEPIVOTCANDIDATE(t)					{(t) = (t) & 0x7FFFFFFF;}
// --- BIT OPERATIONS TO BE PERFORMED ON D_TRANS ENTRIES
#define LOCK(t)									{(t) = (t) | 0x80000000;}
#define ISLOCKED(a, t)							{(a) = (((t) >> 31) > 0) ? 1 : 0;}
#define HASLOCK(t)								(((t) >> 31) > 0)
#define UNLOCK(t)								{(t) = (t) & 0x7FFFFFFF;}
#define LOCKED(t)								((t) | 0x80000000)
#define STRIPLOCK(t)							((t) & 0x7FFFFFFF)
#define STRIPLOCKS(t)							((t) & 0x3FFFFFFF)
// --- BIT OPERATIONS TO BE PERFORMED ON D_TRANS_LABELS ENTRIES TO MARK TAU-TRANSITIONS AS VISIBLE
#define SETVISIBLE(t)							{(t) = (t) | 0x40000000;}
#define SETINVISIBLE(t)							{(t) = (t) & 0xBFFFFFFF;}
#define ISVISIBLE(a, t)							{(a) = (((t) & 0x40000000) > 0) ? 1 : 0;}

// --- BIT OPERATIONS FOR WRITE_STATUS IN COMPARING_SIGNATURES KERNEL
// LOCKS PRESENT:
// - 1. WRITE OCCURRED
// - 2. LOCK CANDIDATES
// - 3. MATCHING PIVOT FOUND
// - 4. STATE PROCESSED
// - 5. INTERMEDIATE COMPARISON RESULTS HAVE BEEN WRITTEN TO GLOBAL MEMORY (TRANSITIONS)
#define CURRENT_WRITE_OCCURRED(t)				((t) & 1)
#define SET_CURRENT_WRITE_OCCURRED(t)			{(t) = (t) | 1;}
#define RESET_CURRENT_WRITE_OCCURRED(t)			{(t) = (t) & 0xFFFFFFFE;}
#define ISLOCKED_CANDIDATES(a, t)				{(a) = (((t) & 2) > 0) ? 1 : 0;}
#define LOCK_CANDIDATES(t)						{(t) = (t) | 2;}
#define UNLOCK_CANDIDATES(t)					{(t) = (t) & 0xFFFFFFFD;}
#define IS_MATCHING_PIVOT_FOUND(a, t)			{(a) = (((t) & 4) > 1) ? 1 : 0;}
#define SET_MATCHING_PIVOT_FOUND(t)				{(t) = (t) | 4;}
#define IS_STATE_PROCESSED(a, t)				{(a) = (((t) & 8) > 1) ? 1 : 0;}
#define SET_STATE_PROCESSED(t)					{(t) = (t) | 8;}
#define RESET_STATE_PROCESSED(t)				{(t) = (t) & 0xFFFFFFF7;}
#define GLOBAL_MEM_WRITE_OCCURRED(a, t)			((a) = (((t) & 16) > 1) ? 1 : 0)
#define SET_GLOBAL_MEM_WRITE_OCCURRED(t)		{(t) = (t) | 16;}
#define RESET_GLOBAL_MEM_WRITE_OCCURRED(t)		{(t) = (t) & 0xFFFFFFEF;}

#define THREADID								threadIdx.x
#define GLOBAL_THREADID							(blockIdx.x * blockDim.x) + threadIdx.x
#define GLOBAL_WARPID							(blockIdx.x * (blockDim.x / WARP_SIZE)) + (threadIdx.x / WARP_SIZE)
#define WARPID									threadIdx.x / WARP_SIZE
#define WARPTID									threadIdx.x % WARP_SIZE
#define WARPLEADERID							((threadIdx.x / WARP_SIZE) * WARP_SIZE)
//#define GLOBAL_LASTINWARP_TID					GLOBAL_THREADID + (31 - WARPTID)
#define NROFTHREADS								gridDim.x * blockDim.x
#define SEARCH_NROFTHREADS						((NROFTHREADS) / 2)
#define SEARCH_THREADID							(GLOBAL_THREADID) % (SEARCH_NROFTHREADS)
//#define SEARCH_NROFTHREADS						(gridDim.x * blockDim.x) / 2

// macros for broadcasting in warp
#define BROADCAST_IF_WARPED						{	warpcont = thcont; \
													if (STEP_SIZE == WARP_SIZE) { \
														shared[THREADID] = thcont; \
														if (warpcont == 0) { \
															for (wj = WARPLEADERID; wj < WARPLEADERID + WARP_SIZE; wj++) { \
																warpcont = shared[wj]; \
																if (warpcont == 1) { \
																	break; \
																} \
															} \
														} \
													} \
												}
#define BROADCAST								{	warpcont = thcont; \
													shared[THREADID] = thcont; \
													if (warpcont == 0) { \
														for (wj = WARPLEADERID; wj < WARPLEADERID + WARP_SIZE; wj++) { \
															warpcont = shared[wj]; \
															if (warpcont == 1) { \
																break; \
															} \
														} \
													} \
												}
#define BREAKWARP								{	if (warpcont == 0) { \
														break; \
													} \
												}

// definitions for signature comparisons
#define STEP_SIZE						((i < d_states_endofsegments) ? WARP_SIZE : 1)
#define PIVOT_STEP_SIZE					((candidates_buffer[ci] < d_states_endofsegments) ? WARP_SIZE : 1)
#define NR_OF_I_BUFFERS					(((trans_end - trans_start)/STEP_SIZE) + (THREAD_BUFFER_SIZE-1)) / THREAD_BUFFER_SIZE
#define NR_OF_PIVOT_BUFFERS				(((pivot_trans_end - pivot_trans_start)/PIVOT_STEP_SIZE) + (THREAD_BUFFER_SIZE-1)) / THREAD_BUFFER_SIZE
#define NOT_RELEVANT					2
#define STILL_POSSIBLE					3
#define SIGNATURES_EQUAL				(signatures_are_equal == 1 || signatures_are_equal == STILL_POSSIBLE)
#define TAU								0
#define VISIBLE_TAU						0x40000000
// macro to enable/disable smart pivot candidate selection depending on label of first outgoing transition
//#define LABELDEPENDENCY					label

// definitions for block
#define EMPTY_CONFBLOCK					0x80000000
#define NA_CONFBLOCK					0xFFFFFFFF

#define LABELDEPENDENCY					0

// macros for broadcasting in warp; shfl version (CUDA 3)
/* #define BROADCAST_IF_WARPED						{	warpcont = thcont; \
													if (STEP_SIZE == WARP_SIZE) { \
														for (j = 0; j < WARP_SIZE; j++) { \
															warpcont |= (uint32_t) __shfl((int) thcont, (int) j); \
														} \
													} \
												}
#define BROADCAST								{	warpcont = thcont; \
													for (j = 0; j < WARP_SIZE; j++) { \
														warpcont |= (uint32_t) __shfl((int) thcont, (int) j); \
													} \
												} */

// start indices of the different parts of the data
#define s_cont				0
#define s_cont2				1
#define s_cont_closure		2
#define s_cont_pivot		3
#define s_offsets			4
#define s_trans_labels		nstates+5
#define s_trans_tgts		nstates+5+trans_cnt
#define s_results			nstates+5+2*trans_cnt

#define ds_cont				0
#define ds_cont2			1
#define ds_cont_closure		2
#define ds_cont_pivot		3
#define ds_offsets			4
#define ds_trans_labels		d_nstates+5
#define ds_trans_tgts		d_nstates+5+d_ntrans
#define ds_results			d_nstates+5+2*d_ntrans

#define TAU_CLOSURE		1
#define NO_TAU_CLOSURE	0

/**
 * This macro checks return value of the CUDA runtime call and exits
 * the application if the call failed.
 */
#define CUDA_CHECK_RETURN(value) {											\
	cudaError_t _m_cudaStat = value;										\
	if (_m_cudaStat != cudaSuccess) {										\
		fprintf(stderr, "Error %s at line %d in file %s\n",					\
				cudaGetErrorString(_m_cudaStat), __LINE__, __FILE__);		\
		exit(1);															\
	} }

// GPU CONSTANTS
__constant__ uint32_t d_nstates;
__constant__ uint32_t d_ntrans;
__constant__ uint32_t d_nlabels;
__constant__ uint32_t d_states_endofsegments;
__constant__ uint32_t d_trans_segmentinterval;

// TEST FUNCTIONS

// hashable struct to count the number of states in the reduced LTS
typedef struct statenode {
	int id;
	UT_hash_handle hh;
} statenode;

// hashtable
statenode *stateset;
uint32_t statesetcount = 0;

// clean the hashtable
void clean_stateset() {
	statenode *current, *tmp;
	
	HASH_ITER(hh, stateset, current, tmp) {
		HASH_DEL(stateset, current);
		free(current);
	}
	statesetcount = 0;
}

// print the hashtable content
void print_stateset() {
	statenode *current, *tmp;
	
	HASH_ITER(hh, stateset, current, tmp) {
		fprintf (stdout, "%d ", current->id);
	}
	fprintf (stdout, "\n");
}

// print the blocks
void print_blocks(uint32_t *tst_results, uint32_t nstates) {
	statenode *current, *tmp;
	uint32_t j;
	
	HASH_ITER(hh, stateset, current, tmp) {
		fprintf (stdout, "[%d]: ", current->id);
		for (j = 0; j < nstates; j++) {
			if (tst_results[j] == current->id) {
				fprintf (stdout, "%d ", j);
			}
		}
		fprintf(stdout, "\n");
	}
	fprintf (stdout, "\n");
}

// add a state to the stateset;
void addstate(uint32_t id) {
	statenode *s;
	int statecast = (int) id;
	HASH_FIND_INT(stateset, &statecast, s);
	if (s==NULL) {
		s = (statenode*) malloc(sizeof(statenode));
		s->id = statecast;
		HASH_ADD_INT(stateset, id, s);
		statesetcount++;
	}
}

int vmem = 0;

//wrapper around cudaMalloc to count allocated memory and check for error while allocating
int cudaMallocCount ( void ** ptr,int size) {
	cudaError_t err = cudaSuccess;
	vmem += size;
	err = cudaMalloc(ptr,size);
	if (err) {
		printf("Error %s at line %d in file %s\n", cudaGetErrorString(err), __LINE__, __FILE__);
		exit(1);
	}
	fprintf (stdout, "allocated %d\n", size);
	return size;
}

// test function, printing the invisible states
void print_invisiblestates(uint32_t *d_results, uint32_t nstates) {
	uint32_t result;
	uint32_t *tst_results = (uint32_t *) malloc(nstates * sizeof(uint32_t));
	CUDA_CHECK_RETURN(cudaMemcpy(tst_results, d_results, nstates * sizeof(uint32_t), cudaMemcpyDeviceToHost));

	clean_stateset();
	for (int i = 0; i < nstates; i++) {
		ISLOCKED(result, tst_results[i]);
		if (result == 1) {
			addstate(STRIPLOCK(tst_results[i]));
		}
	}
	fprintf (stdout, "invisible states: %d\n", statesetcount);
	print_stateset();
	//print_blocks(tst_results, nstates);
	clean_stateset();
}

// test function, printing the number of states in the output LTS on the screen
void print_outputsize(uint32_t *d_results, uint32_t nstates) {
	uint32_t *tst_results = (uint32_t *) malloc(nstates * sizeof(uint32_t));
	CUDA_CHECK_RETURN(cudaMemcpy(tst_results, d_results, nstates * sizeof(uint32_t), cudaMemcpyDeviceToHost));

	clean_stateset();
	for (int i = 0; i < nstates; i++) {
		if (STRIPLOCKS(tst_results[i]) != NOBLOCK) {
			addstate(STRIPLOCKS(tst_results[i]));
		}
	}
	fprintf (stdout, "nr of states in reduced LTS: %d\n", statesetcount);
	//print_stateset();
	//print_blocks(tst_results, nstates);
	clean_stateset();
	free(tst_results);
}

// test function, printing the content of the global memory on the screen
void print_status(uint32_t *d_offsets, uint32_t *d_trans_labels, uint32_t *d_trans_tgts, uint32_t *d_results, uint32_t nstates, uint32_t ntrans) {
	uint32_t *tst_offsets = (uint32_t *) malloc((nstates+1) * sizeof(uint32_t));
	uint32_t *tst_trans_labels = (uint32_t *) malloc(ntrans * sizeof(uint32_t));
	uint32_t *tst_trans_tgts = (uint32_t *) malloc(ntrans * sizeof(uint32_t));
	uint32_t *tst_results = (uint32_t *) malloc(nstates * sizeof(uint32_t));
	uint32_t result;

	CUDA_CHECK_RETURN(cudaMemcpy(tst_offsets, d_offsets, (nstates+1) * sizeof(uint32_t), cudaMemcpyDeviceToHost));
	CUDA_CHECK_RETURN(cudaMemcpy(tst_trans_labels, d_trans_labels, ntrans * sizeof(uint32_t), cudaMemcpyDeviceToHost));
	CUDA_CHECK_RETURN(cudaMemcpy(tst_trans_tgts, d_trans_tgts, ntrans * sizeof(uint32_t), cudaMemcpyDeviceToHost));
	CUDA_CHECK_RETURN(cudaMemcpy(tst_results, d_results, nstates * sizeof(uint32_t), cudaMemcpyDeviceToHost));
	
	fprintf (stdout, "offsets (with lock info):\n");
	for (int i = 0; i < nstates+1; i++) {
		ISLOCKED(result, tst_offsets[i]);
		fprintf(stdout, "%d: (%d,%d)  ", i, result, STRIPLOCK(tst_offsets[i]));
		fprintf (stdout, "\n");
	}
	fprintf (stdout, "trans labels (with lock info):\n");
	for (int i = 0; i < ntrans; i++) {
		ISLOCKED(result, tst_trans_labels[i]);
		fprintf(stdout, "%d: (%d,%d)  ", i, result, STRIPLOCK(tst_trans_labels[i]));
		fprintf (stdout, "\n");
	}
	fprintf (stdout, "trans targets (with lock info):\n");
	for (int i = 0; i < ntrans; i++) {
		ISLOCKED(result, tst_trans_tgts[i]);
		fprintf(stdout, "%d: (%d,%d)  ", i, result, STRIPLOCK(tst_trans_tgts[i]));
		fprintf (stdout, "\n");
	}
	fprintf (stdout, "results (with lock info):\n");
	for (int i = 0; i < nstates; i++) {
		ISLOCKED(result, tst_results[i]);
		fprintf(stdout, "%d: (%d,%d)  ", i, result, STRIPLOCKS(tst_results[i]));
		fprintf (stdout, "\n");
	}
	free(tst_offsets);
	free(tst_trans_labels);
	free(tst_trans_tgts);
	free(tst_results);
}

// test function, printing the content of the global memory on the screen
void print_results(uint32_t *d_results, uint32_t nstates) {
	uint32_t *tst_results = (uint32_t *) malloc(nstates * sizeof(uint32_t));
	uint32_t result;

	CUDA_CHECK_RETURN(cudaMemcpy(tst_results, d_results, nstates * sizeof(uint32_t), cudaMemcpyDeviceToHost));
	
	fprintf (stdout, "results (with lock info):\n");
	for (int i = 0; i < nstates; i++) {
		ISLOCKED(result, tst_results[i]);
		fprintf(stdout, "%d: (%d,%d)  ", i, result, STRIPLOCKS(tst_results[i]));
		fprintf (stdout, "\n");
	}
}

// test function, printing the content of the global memory on the screen
void print_signatures(uint32_t *d_offsets, uint32_t *d_trans_labels, uint32_t *d_trans_tgts, uint32_t *d_results, uint32_t nstates, uint32_t ntrans) {
	uint32_t *tst_offsets = (uint32_t *) malloc((nstates+1) * sizeof(uint32_t));
	uint32_t *tst_trans_labels = (uint32_t *) malloc(ntrans * sizeof(uint32_t));
	uint32_t *tst_trans_tgts = (uint32_t *) malloc(ntrans * sizeof(uint32_t));
	uint32_t *tst_results = (uint32_t *) malloc(nstates * sizeof(uint32_t));
	uint32_t result, result2;

	CUDA_CHECK_RETURN(cudaMemcpy(tst_offsets, d_offsets, (nstates+1) * sizeof(uint32_t), cudaMemcpyDeviceToHost));
	CUDA_CHECK_RETURN(cudaMemcpy(tst_trans_labels, d_trans_labels, ntrans * sizeof(uint32_t), cudaMemcpyDeviceToHost));
	CUDA_CHECK_RETURN(cudaMemcpy(tst_trans_tgts, d_trans_tgts, ntrans * sizeof(uint32_t), cudaMemcpyDeviceToHost));
	CUDA_CHECK_RETURN(cudaMemcpy(tst_results, d_results, nstates * sizeof(uint32_t), cudaMemcpyDeviceToHost));
	
	for (uint32_t i = 0; i < MIN(nstates,100); i++) {
		ISLOCKED(result, tst_results[i]);
		ISLOCKED(result2, tst_offsets[i]);
		uint32_t offsetstart = STRIPLOCK(tst_offsets[i]);
		uint32_t offsetend = STRIPLOCK(tst_offsets[i+1]);
		uint32_t from_region = STRIPLOCKS(tst_results[i]);
		for (uint32_t j = offsetstart; j < offsetend; j++) {
			uint32_t label = STRIPLOCK(tst_trans_labels[j]);
			uint32_t to = STRIPLOCK(tst_trans_tgts[j]);
			uint32_t to_region = STRIPLOCKS(tst_results[to]);
			fprintf (stdout, "%d(%d,%d,%d): %d->%d (%d)\n", i, from_region, result, result2, label, to, to_region);
		}
	}
	free(tst_offsets);
	free(tst_trans_labels);
	free(tst_trans_tgts);
	free(tst_results);
}

// function to set all values in d_results to an initial value
__global__ void init_results(uint32_t *d_data) {
	for (uint32_t i = GLOBAL_THREADID; i < d_nstates; i += NROFTHREADS) {
		*(d_data+ds_results+i) = NOBLOCK;
	}
}

// Set all states as candidates for block representation (strong bisimulation reduction).
__global__ void identify_candidates_sb(uint32_t *d_offsets) {
	uint32_t i, result;
	
	for (i = GLOBAL_THREADID; i < d_nstates; i += NROFTHREADS) {
		result = d_offsets[i];
		LOCK(result);
		d_offsets[i] = result;
	}
}

// Set all states that do not have an invisible tau-transition as candidates for block representation (branching bisimulation reduction).
__global__ void identify_candidates_bb(uint32_t *d_offsets, uint32_t *d_results) {
	extern volatile __shared__ uint32_t shared[];

	uint32_t i, wj, result, block, thcont, warpcont;
	
	for (i = GLOBAL_THREADID; i < d_nstates; i += NROFTHREADS) {
		result = d_offsets[i];
		block = d_results[i];
		BLOCKISPROPAGATED(thcont, block);
		if (thcont == 0) {
			LOCK(result);
			thcont = 1;
		}
		else {
			thcont = 0;
		}
		BROADCAST;
		if (warpcont == 1) {
			d_offsets[i] = result;
		}
	}
}

//// Set all states as candidates for block representation, and mark states for closure checking
//// (branching bisimulation reduction).
//__global__ void identify_candidates_bb(uint32_t *d_offsets, uint32_t *d_trans_labels, uint32_t *d_results) {
//	extern volatile __shared__ uint32_t shared[];
//	
//	uint32_t i, wj, result, trans_start, trans_end, trans, thcont, warpcont, block;
//	
//	for (i = GLOBAL_THREADID; i < d_nstates; i += NROFTHREADS) {
//		result = d_offsets[i];
//		LOCK(result);
//		d_offsets[i] = result;
//		
//		GETOFFSET(trans_start, result);
//		if (STEP_SIZE == WARP_SIZE) {
//			trans_end = trans_start+(WARP_SIZE*d_trans_segmentinterval);
//		}
//		else {
//			GETOFFSET(trans_end, d_offsets[i+1]);
//		}
//		thcont = 0;
//		if (trans_end - trans_start > 0) {
//			// check label of first transition
//			trans =	d_trans_labels[trans_start];
//			if (trans == TAU) {
//				// there is a tau-transition, mark this state
//				thcont = 1;
//			}
//		}
//		BROADCAST;
//		if (warpcont == 1) {
//			block = d_results[i];
//			if (thcont == 1) {
//				SETCLOSURECHECK(block);
//			}
//			d_results[i] = block;			
//		}
//	}
//}

// Set all block results of pivot candidates states to NOBLOCK.
__global__ void reset_blocks(uint32_t *d_results) {
	extern volatile __shared__ uint32_t shared[];
	
	uint32_t wj, i, block, result, thcont, warpcont;
	
	for (i = GLOBAL_THREADID; i < d_nstates; i += NROFTHREADS) {
		block = d_results[i];
		thcont = 0;
		ISPIVOTCANDIDATE(result, block);
		if (result == 1) {
			thcont = 1;
		}
		BROADCAST;
		if (warpcont == 1) {
			d_results[i] = NOBLOCK;
		}
	}
}

//// Mark which states have tau-transitions,
//// and therefore also require closure checking (branching bisimulation reduction).
//__global__ void identify_closure(uint32_t *d_offsets, uint32_t *d_trans_labels, uint32_t *d_results) {
//	extern volatile __shared__ uint32_t shared[];
//
//	uint32_t i, wj, trans_start, trans_end, trans, block, thcont, warpcont;
//	
//	for (i = GLOBAL_THREADID; i < d_nstates; i += NROFTHREADS) {
//		GETOFFSET(trans_start, d_offsets[i]);
//		if (STEP_SIZE == WARP_SIZE) {
//			trans_end = trans_start+(WARP_SIZE*d_trans_segmentinterval);
//		}
//		else {
//			GETOFFSET(trans_end, d_offsets[i+1]);
//		}
//		thcont = 0;
//		if (trans_end - trans_start > 0) {
//			// check label of first transition
//			trans =	d_trans_labels[trans_start];
//			if (trans == TAU) {
//				// there is a tau-transition, mark this state
//				thcont = 1;
//			}
//		}
//		BROADCAST;
//		if (warpcont == 1) {
//			block = d_results[i];
//			if (thcont == 1) {
//				SETCLOSURECHECK(block);
//			}
//			d_results[i] = block;			
//		}
//	}
//}

//// Identify new candidates for block representation at initialisation of reduction. Also indicate in d_results which states require closure checking.
//// Only needed for branching bisimulation reduction. For strong, every state can represent.
//__global__ void identify_candidates_bb(uint32_t *d_offsets, uint32_t *d_trans_labels, uint32_t *d_trans_tgts, uint32_t *d_results) {
//	extern volatile __shared__ uint32_t shared[];
//
//	uint32_t i, j, wj, current_state, result, trans_start, trans_end, trans, thcont, warpcont;
//
//	for (i = GLOBAL_THREADID; i < d_nstates; i += NROFTHREADS) {
//		current_state = d_results[i];
//		GETOFFSET(trans_start, d_offsets[i]);
//		if (STEP_SIZE == WARP_SIZE) {
//			trans_end = trans_start+(WARP_SIZE*d_trans_segmentinterval);
//		}
//		else {
//			GETOFFSET(trans_end, d_offsets[i+1]);
//		}
//		result = 1;
//		thcont = 1;
//		warpcont = 1;
//		for (j = trans_start; j < trans_end && warpcont == 1; j += STEP_SIZE) {
//			trans = d_trans_labels[j];
//			// if it is a tau-transition, and we do branching bisim reduction, closure checking is required for this state
//			if (thcont == 1) {
//				if (trans == TAU) {
//					// there is a tau-transition. Not suitable as pivot.
//					thcont = 0;
//					result = 0;
//				}
//				else {
//					// we have seen all tau-transitions, and can stop now.
//					thcont = 0;
//				}
//			}
//			BROADCAST_IF_WARPED;
//		}
//		// result == 1 indicates that state can be pivot
//		if (result == 1) {
//			LOCK(trans_start);
//			thcont = 1;
//		}
//		BROADCAST;
//		if (warpcont == 1) {
//			d_offsets[i] = trans_start;
//		}
//		thcont = 0;
//		// state i needs closure checking
//		if (result == 0) {
//			SETCLOSURECHECK(current_state);
//			thcont = 1;
//		}
//		BROADCAST;
//		if (warpcont == 1) {
//			d_results[i] = current_state;
//		}
//	}
//}

// Choose new representatives of discovered blocks
// labelnr indicates which outgoing label to consider for hashing
// iter is used to determine the boundaries of the current hashing round
__global__ void choose_representatives(uint32_t *d_offsets, uint32_t *d_trans_labels, uint32_t *d_trans_tgts, uint32_t *d_results, uint32_t *d_cont, uint32_t *d_cont2, uint32_t labelnr, uint32_t iter) {
	extern volatile __shared__ uint32_t shared[];

	uint32_t i, j, wj, k, block, hash, result, trans_start, trans_end, trans, label, thcont, thconttmp, warpcont;

	for (i = GLOBAL_THREADID; i < d_nstates; i += NROFTHREADS) {
		thcont = 0;
		trans_start = d_offsets[i];
		ISPIVOTCANDIDATE(result, trans_start);
		if (result == 1) {
			thcont = 1;
		}
		BROADCAST_IF_WARPED;
		if (warpcont == 1) {
			GETOFFSET(trans_start, trans_start);
			if (STEP_SIZE == WARP_SIZE) {
				trans_end = trans_start+(WARP_SIZE*d_trans_segmentinterval);
			}
			else {
				GETOFFSET(trans_end, d_offsets[i+1]);
			}
			// fetch the labelnr-th action label in the list of outgoing transitions for hashing
			label = 0;
			// temporarily store earlier thcont result
			thconttmp = thcont;
			if (thcont == 1) {
				if (((trans_end - trans_start) / STEP_SIZE) <= labelnr) {
					thcont = 0;
				}
			}
			BROADCAST_IF_WARPED;
			if (warpcont == 1) {
				// fetch the right label
				label = d_nlabels;
				k = 0;
				//PRINTTHREAD(i, trans_end - trans_start);
				for (j = trans_start; j < trans_end && warpcont == 1; j += STEP_SIZE) {
					result = d_trans_labels[j];
					if (thcont == 1) {
						if (result != EMPTYTRANS && result != TAU) {
							if (STRIPLOCKS(result) == TAU) {
								result = 0;
							}
							if (result != label) {
								label = result;
								if (k == labelnr) {
									thcont = 0;
								}
								else {
									k++;
								}
							}
						}
					}
					BROADCAST_IF_WARPED;
				}
				// if thcont is still 1, then a suitable label was not found
				if (thcont == 1) {
					label = 0;
				}
			}
			// set thcont back to earlier result
			thcont = thconttmp;
			block = d_results[i];
			// thcont indicates if state can be pivot
			if (thcont == 1) {
				GETREPRESENTATIVE(block, block);
				if (block == NOBLOCK) {
					block = 0;
				}
				//PRINTTHREAD(5, i);
				//PRINTTHREAD(0, block);
				//PRINTTHREAD(3, label);
				hash = block*d_nlabels + LABELDEPENDENCY;
				//PRINTTHREAD(block, LABELDEPENDENCY);
				// is the hash valid for the current selection iteration?
				if ((hash / d_ntrans) == iter) {
					// try to claim representative status
					hash = hash % d_ntrans;
					trans = d_trans_tgts[hash];
					ISLOCKED(result, trans);
					if (result == 0) {
						block = i;
						LOCK(block);
						result = atomicCAS((uint32_t *) d_trans_tgts+hash, trans, block);
						if (trans == result) {
							// we were successful in claiming representative status
							// temporarily copy trans entry in d_results
							//PRINTTHREAD(2, result);
							// indicate that after this iteration, postprocessing is required
							(*d_cont2) = 1;
						}
						else {
							// we did not succeed. take over the new representative info.
							UNLOCK(result);
							//PRINTTHREAD(4, result);
						}
					}
					else {
						// we did not succeed. take over the new representative info.
						result = STRIPLOCK(trans);
						//PRINTTHREAD(4, result);
					}
				}
				else {
					// we have an entry that could not be handled yet. we leave trans_start locked, indicating that a next iteration is required
					result = block;
					MAKEPIVOTCANDIDATE(trans_start);
					thcont = 0;
					(*d_cont) = 1;
				}
			}
			else {
				result = block;
			}
			BROADCAST_IF_WARPED;
			if (warpcont == 1) {
				d_results[i] = result;
				d_offsets[i] = trans_start;
			}
		}
	}
}

// postprocess the results of the last representative selection iteration
__global__ void postprocess_representatives(uint32_t *d_offsets, uint32_t *d_trans_tgts, uint32_t *d_results) {
	uint32_t i, entry, state;

	for (i = GLOBAL_THREADID; i < d_ntrans; i += NROFTHREADS) {
		entry = d_trans_tgts[i];
		ISLOCKED(state, entry);
		if (state == 1) {
			// we need to replace this again with the right trans entry
			GETREPRESENTATIVE(state, entry);
			d_trans_tgts[i] = d_results[state];
			// set pivot as itself
			d_results[state] = state;
		}
	}
}

// Mark tau-transitions between two different blocks as visible.
__global__ void detect_visible_taus(uint32_t *d_offsets, uint32_t *d_trans_labels, uint32_t *d_trans_tgts, uint32_t *d_results) {
	extern volatile __shared__ uint32_t shared[];
	
	uint32_t wj, i, j, trans_start, trans_end, label, tgt, block_src, block_tgt, thcont, warpcont;
	
	for (i = GLOBAL_THREADID; i < d_nstates; i += NROFTHREADS) {
		//nr_of_invisible_taus = 0;
		GETOFFSET(trans_start, d_offsets[i]);
		block_src = d_results[i];
		if (STEP_SIZE == WARP_SIZE) {
			trans_end = trans_start+(WARP_SIZE*d_trans_segmentinterval);
		}
		else {
			GETOFFSET(trans_end, d_offsets[i+1]);
		}
		thcont = 1;
		warpcont = 1;
		for (j = trans_start; j < trans_end && warpcont == 1; j += STEP_SIZE) {
			label = d_trans_labels[j];
			tgt = d_trans_tgts[j];
			if (thcont == 1) {
				thcont = 0;
				if (label == TAU) {
					GETREPRESENTATIVE(block_tgt, d_results[tgt]);
					if (STRIPLOCK(block_src) != block_tgt) {
						label = VISIBLE_TAU;
						thcont = 1;
					}
					//else {
					//	nr_of_invisible_taus++;
					//}
				}
				else if (label == VISIBLE_TAU) {
					GETREPRESENTATIVE(block_tgt, d_results[tgt]);
					if (STRIPLOCK(block_src) == block_tgt) {
						label = TAU;
						thcont = 1;
						//nr_of_invisible_taus++;
					}
				}
			}
			BROADCAST_IF_WARPED;
			if (warpcont == 1) {
				d_trans_labels[j] = label;
			}
			thcont = (STRIPLOCKS(label) == TAU) ? 1 : 0;
			BROADCAST_IF_WARPED;
		}
		// if the number of invisible taus is zero, but the state is marked for closure checking, uncheck it
		//thcont = 0;
		//if (block_src == LOCKED(block_src) && nr_of_invisible_taus == 0) {
		//	UNLOCK(block_src);
		//	thcont = 1;
		//}
		//BROADCAST;
		//if (warpcont == 1) {
		//	d_results[i] = block_src;
		//}
	}
}

//// if d_cont2 is true, check if states need to be moved to other block. This is the case if a previous
//// pivot has moved blocks due to the detection of an outgoing invisible tau-transition
//__global__ void postprocess_tauclosures(uint32_t *d_results) {
//	extern volatile __shared__ uint32_t shared[];
//	
//	uint32_t wj, i, block_src, block_pivot, result, thcont, warpcont;
//	
//	for (i = GLOBAL_THREADID; i < d_nstates; i += NROFTHREADS) {
//		thcont = 0;
//		GETREPRESENTATIVE(block_src, d_results[i]);
//		block_pivot = d_results[block_src];
//		BLOCKISPROPAGATED(result, block_pivot);
//		if (result == 1) {
//			GETREPRESENTATIVE(block_src, block_pivot);
//			thcont = 1;
//		}
//		BROADCAST;
//		if (warpcont == 1) {
//			d_results[i] = block_src;
//		}
//	}
//}

// compare signatures of states with those of their pivot
// LOCKING:
// - In local buffers:
//    - If buffer_labels entry is locked, then match has been found in buffer(s) of other state
//    - If buffer_tgts entry is locked, then entry is actually the block that the tgt state belongs to
// - In global memory:
//    - If d_trans_tgts entry is locked, then match has been found for state i of a transition of pivot state
//	FLAG D_CONT IS USED AS FOLLOWS: d_cont is set to true if there are new pivot candidates for next round, in closure checking, d_cont is set if tau-closure needs another iteration
//  FLAG D_CONT2 IS USED: to indicate that a pivot has been identified as invisible (has an invisible tau transition). Then, when postprocessing tau-closure, we move all states represented
//  by the pivot to the same block
__global__ void compare_signatures(uint32_t *d_offsets, uint32_t *d_trans_labels, uint32_t *d_trans_tgts, uint32_t *d_results, uint32_t *d_cont, uint32_t *d_cont2, uint32_t tauclosure) {
	extern volatile __shared__ uint32_t shared[];

	uint32_t i, j, wj, k, pi, bii, bpi, ii, ci, result, trans_start, trans_end, pivot_trans_start, pivot_trans_end, block, thcont, thcont2, warpcont;
	// write_status combines two things: 1) has a write occurred or not for the current i buffer? (1 or 0), 2) has a write occurred for at least one of the i buffers? (locked or not)
	uint32_t signatures_are_equal, write_status = 0;
	// buffer for storing transition info of state i
	uint32_t i_buffer_labels[THREAD_BUFFER_SIZE];
	uint32_t i_buffer_tgts[THREAD_BUFFER_SIZE];
	uint32_t i_buffer_size = 0;
	uint32_t i_elements_checked = 0;
	uint32_t comparing_current = 0;
	// buffer for storing transition info of pivot
	uint32_t pivot_buffer_labels[THREAD_BUFFER_SIZE];
	uint32_t pivot_buffer_tgts[THREAD_BUFFER_SIZE];
	uint32_t pivot_buffer_size = 0;
	// buffer for storing pivot candidates
	uint32_t candidates_buffer[THREAD_BUFFER_SIZE];
	uint32_t candidates_buffer_size;
	// points to transition to possible next candidate when checking tau-closure (to take care of overflowing candidates buffer)
	uint32_t next_candidate;
	// closure: suitable target block found to propagate, and block of current tgt
	uint32_t conf_block, tgt_block;

	for (i = GLOBAL_THREADID; i < d_nstates; i += NROFTHREADS) {
		//complete_signature = 1;
		candidates_buffer_size = 0;
		thcont = 0;
		write_status = 0;
		block = d_results[i];
		conf_block = EMPTY_CONFBLOCK;
		signatures_are_equal = NOT_RELEVANT;
		// check if state is not pivot when doing strong comparisons
		if (STRIPLOCK(block) != i || tauclosure == TAU_CLOSURE) {
			// check whether state has a propagated block value
			BLOCKISPROPAGATED(result, block);
			GETREPRESENTATIVE(candidates_buffer[candidates_buffer_size], block);
			// two possible conditions, depending on whether we check tau-closure or not
			//if (tauclosure == NO_TAU_CLOSURE || result == 1) {
				// compare to signature of pivot
				// this thread needs to read signature of its state. Warp will go along with this
			if (tauclosure == TAU_CLOSURE || result == 0) {
				candidates_buffer_size++;
				// else: checking for tau-closure
				// next, outside of this if-clause, we will check (possibly in a warped fashion), whether state i has
				// a confluent tau-transition.
				thcont = 1;
			}
			//}
		}
		BROADCAST;
		if (warpcont == 1) {
			// get end offset of state i
			trans_start = d_offsets[i];
			GETOFFSET(trans_start, trans_start);
			if (STEP_SIZE == WARP_SIZE) {
				trans_end = trans_start+(WARP_SIZE*d_trans_segmentinterval);
			}
			else {
				GETOFFSET(trans_end, d_offsets[i+1]);
			}
			if (tauclosure == 1) {
				next_candidate = trans_start;
			}
		}
		// Start of while loop for candidate buffer (candidates for blocks to which state i may belong)
		while (warpcont == 1) {
			if (tauclosure == 1) {
				candidates_buffer_size = 0;
				// obtain the target states reachable via tau-transitions
				for (j = next_candidate; j < trans_end && warpcont == 1; j += STEP_SIZE) {
					ii = d_trans_labels[j];
					result = d_trans_tgts[j];
					if (thcont == 1) {
						if (STRIPLOCKS(ii) == TAU) {
							//if (STRIPLOCK(ii) == TAU) {
							//	inv_tau_count++;
							//}
							pi = d_results[result];
							BLOCKISPROPAGATED(ci, pi);
							if (ci == 1 && pi != LOCKED_NOBLOCK) {
								result = STRIPLOCK(pi);
							}
							for (pi = 0; pi < candidates_buffer_size; pi++) {
								//if (result == STRIPLOCK(candidates_buffer[pi])) {
								if (result == candidates_buffer[pi]) {
									break;
								}
							}
							if (pi == candidates_buffer_size) {
								if (pi == THREAD_BUFFER_SIZE) {
									// we have another candidate, but buffer is full. keep a pointer to it for a next iteration.
									// if this is not the first iteration, then we don't need to check whether the signature is complete,
									// and we can stop analysing the outgoing transitions
									next_candidate = j;
									LOCK_CANDIDATES(write_status);
									if (next_candidate != trans_start) {
										thcont = 0;
									}
								}
								else {
									//if (ii == VISIBLE_TAU) {
									//	LOCK(result);
									//}
									//candidates_label_buffer[candidates_buffer_size] = ii;
									candidates_buffer[candidates_buffer_size++] = result;
								}
							}
							//else {
							//	if (ii == VISIBLE_TAU) {
							//		LOCK(candidates_buffer[pi]);
							//	}
							//}
						}
						else {
							// due to label-sorted transitions, we know that no more candidates can come up.
							thcont = 0;
						}
					}
					BROADCAST_IF_WARPED;
				}
			}
			// determine if there are pivots to consider in warp.
			// do another while iteration if this is the case.
			ci = 0;
			thcont = (ci < candidates_buffer_size) ? 1 : 0;
			BROADCAST_IF_WARPED;
			// while over candidates in current candidates buffer
			while (warpcont == 1) {
				signatures_are_equal = NOT_RELEVANT;
				// get block of target
				tgt_block = 0;
				if (thcont == 1) {
					tgt_block = (tauclosure == TAU_CLOSURE) ? STRIPLOCK(d_results[candidates_buffer[ci]]) : candidates_buffer[ci];
				}
				// get offsets for pivot, if required
				if (thcont == 1) {
					GETOFFSET(pivot_trans_start, d_offsets[candidates_buffer[ci]]);
					//GETOFFSET(pivot_trans_start, d_offsets[STRIPLOCK(candidates_buffer[ci])]);
					if (PIVOT_STEP_SIZE == WARP_SIZE) {
						pivot_trans_end = pivot_trans_start+(WARP_SIZE*d_trans_segmentinterval);
					}
					else {
						//GETOFFSET(pivot_trans_end, d_offsets[STRIPLOCK(candidates_buffer[ci])+1]);
						GETOFFSET(pivot_trans_end, d_offsets[candidates_buffer[ci]+1]);
					}
					// current transitions of pivot
					k = pivot_trans_start;
					signatures_are_equal = 1;
				}
				// we set first entry in i_buffer_labels to EMPTYTRANS, to avoid repeated reading when
				// a single i buffer is sufficient.
				i_buffer_labels[0] = EMPTYTRANS;
				// range over buffers for pivot state
				bpi = 0;
				while (warpcont == 1) {
					pivot_buffer_size = 0;
					if (thcont == 1) {
						// fill buffer for pivot
						for (; k < pivot_trans_end && pivot_buffer_size < THREAD_BUFFER_SIZE; k += PIVOT_STEP_SIZE) {
							pivot_buffer_labels[pivot_buffer_size] = d_trans_labels[k];
							if (pivot_buffer_labels[pivot_buffer_size] != EMPTYTRANS) {
								// if transition is an invisible TAU, ignore
								if (tauclosure == TAU_CLOSURE || pivot_buffer_labels[pivot_buffer_size] != TAU) {
									pivot_buffer_tgts[pivot_buffer_size] = d_trans_tgts[k];
//									if (GLOBAL_THREADID == 0) {
//										PRINTTHREAD(1, STRIPLOCK(pivot_buffer_labels[pivot_buffer_size]));
//										PRINTTHREAD(2, STRIPLOCK(pivot_buffer_tgts[pivot_buffer_size]));
//										PRINTTHREAD(3, STRIPLOCK(d_results[STRIPLOCK(pivot_buffer_tgts[pivot_buffer_size])]));
//									}
									UNLOCK(pivot_buffer_tgts[pivot_buffer_size]);
									pivot_buffer_size++;
								}
							}
							// we can safely break out of the for loop, since no more transitions are left
							else {
								break;
							}
						}
					}
					// range over buffers for state i
					j = trans_start;
					// willingness to look is stored in thcont2
					thcont2 = thcont;
					bii = 0;
					thcont = (thcont2 == 1 && bii < NR_OF_I_BUFFERS) ? 1 : 0;
					BROADCAST_IF_WARPED;
					while (warpcont == 1) {
						// fill buffer for state i
						thcont = (thcont2 == 1 && (NR_OF_I_BUFFERS > 1 || i_buffer_labels[0] == EMPTYTRANS)) ? 1 : 0;
						BROADCAST_IF_WARPED;
						if (warpcont == 1) {
							i_buffer_size = 0;
							comparing_current = 0;
							i_elements_checked = 0;
							thcont = (thcont == 1 && i_buffer_size < THREAD_BUFFER_SIZE) ? 1 : 0;
							BROADCAST_IF_WARPED;
							for (; j < trans_end && warpcont == 1; j += STEP_SIZE) {
								// get as many transitions as possible or required and place these in local buffer
								result = d_trans_labels[j];
								pi = d_trans_tgts[j];
								if (thcont == 1) {
									i_buffer_labels[i_buffer_size] = result;
									if (i_buffer_labels[i_buffer_size] != EMPTYTRANS) {
										i_buffer_tgts[i_buffer_size] = pi;
//										if (GLOBAL_THREADID == 0) {
//										PRINTTHREAD(4, STRIPLOCK(i_buffer_labels[i_buffer_size]));
//										PRINTTHREAD(5, STRIPLOCK(i_buffer_tgts[i_buffer_size]));
//										PRINTTHREAD(6, STRIPLOCK(d_results[STRIPLOCK(i_buffer_tgts[i_buffer_size])]));
//										}
										UNLOCK(i_buffer_tgts[i_buffer_size]);
										i_buffer_size++;
									}
									i_elements_checked++;
								}
								thcont = (thcont == 1 && i_buffer_size < THREAD_BUFFER_SIZE) ? 1 : 0;
								BROADCAST_IF_WARPED;
							}
						}
						RESET_CURRENT_WRITE_OCCURRED(write_status);
						if (thcont2 == 1) {
							//PRINTTHREAD(1, candidates_buffer[ci]);
							//PRINTTHREAD(2, i_buffer_size);
							// iterate over pivot transitions and compare with transitions of state i. When match found, lock match in buffers
							// (in the labels entries).
							signatures_are_equal = 1;
							for (pi = 0; pi < pivot_buffer_size; pi++) {
								signatures_are_equal = 1;
								for (ii = comparing_current; ii < i_buffer_size; ii++) {
									if (STRIPLOCKS(pivot_buffer_labels[pi]) == STRIPLOCKS(i_buffer_labels[ii])) {
										// compare target regions. If required, fetch block and place in buffers (indicated by locked buffer_tgts entries)
										ISLOCKED(result, i_buffer_tgts[ii]);
										if (result == 0) {
											i_buffer_tgts[ii] = d_results[i_buffer_tgts[ii]];
											LOCK(i_buffer_tgts[ii]);
										}
										ISLOCKED(result, pivot_buffer_tgts[pi]);
										if (result == 0) {
											pivot_buffer_tgts[pi] = d_results[pivot_buffer_tgts[pi]];
											LOCK(pivot_buffer_tgts[pi]);
										}
										if (STRIPLOCK(pivot_buffer_tgts[pi]) == STRIPLOCK(i_buffer_tgts[ii])) {
											// transition is equal. Lock entries
											signatures_are_equal = STILL_POSSIBLE;
											ISLOCKED(result, i_buffer_labels[ii]);
											if (result == 0) {
												// a lock has been placed in the i buffer. If multiple i and pivot buffers are needed, this means that writing
												// to global memory of the current i buffer is needed after the comparison has finished.
												SET_CURRENT_WRITE_OCCURRED(write_status);
												LOCK(i_buffer_labels[ii]);
											}
											LOCK(pivot_buffer_labels[pi]);
										}
									}
									else if (STRIPLOCKS(pivot_buffer_labels[pi]) > STRIPLOCKS(i_buffer_labels[ii])) {
										// shift comparing_current
										comparing_current++;
									}
									else {
										// i transition has a label greater than the pivot transition label. Entry is not present in i buffer beyond this point.
										break;
									}
								}
								// if we are working with a single i buffer, early exit is possible if match not found
								// NOTE: not if we check tau-closure, then we need that pivot can simulate all behaviour of state i, not reverse
								if (tauclosure != TAU_CLOSURE) {
									if (NR_OF_I_BUFFERS < 2 && signatures_are_equal == 1) {
										thcont2 = 0;
										signatures_are_equal = 0;
										break;
									}
								}
							}
						}
						result = thcont;
						thcont = 0;
						// if at least one transition has been locked, and if there are more pivot buffers and multiple i buffers,
						// the current i buffer (labels) must be written to global memory.
						// This can be done safely, since the current thread owns state i.
						if (CURRENT_WRITE_OCCURRED(write_status) == 1 && NR_OF_I_BUFFERS > 1 && bpi+1 < NR_OF_PIVOT_BUFFERS) {
							thcont = 1;
							// indicate that a write is being done
							SET_GLOBAL_MEM_WRITE_OCCURRED(write_status);
						}
//						BROADCAST_IF_WARPED;
//						if (warpcont == 1) {
						if (thcont == 1) {
							ii = 0;
//							for (pi = trans_start; pi < trans_end; pi += STEP_SIZE) {
//								wj = d_trans_labels[pi];
//								if (thcont == 1) {
//									if (pi >= j - (i_elements_checked*STEP_SIZE) && pi < j && wj != EMPTYTRANS) {
//										wj = i_buffer_labels[ii++];
//									}
//								}
//								d_trans_labels[pi] = wj;
//							}
							for (pi = j - (i_elements_checked*STEP_SIZE); pi < j; pi += STEP_SIZE) {
								wj = d_trans_labels[pi];
								if (wj != EMPTYTRANS) {
									d_trans_labels[pi] = i_buffer_labels[ii++];
								}
							}
						}
						thcont = result;
						// if no more comparisons are needed with the current i buffer, check its content for full match
						if (thcont2 == 1) {
							if (NR_OF_PIVOT_BUFFERS == 0 || NR_OF_PIVOT_BUFFERS == bpi+1) {
								for (ii = 0; ii < i_buffer_size; ii++) {
									result = 1;
									if (tauclosure == TAU_CLOSURE) {
										// if this is a tau-transition leading to current pivot, then ignore
										if (STRIPLOCKS(i_buffer_labels[ii]) == TAU) {
											ISLOCKED(result, i_buffer_tgts[ii]);
											if (result == 0) {
												i_buffer_tgts[ii] = d_results[i_buffer_tgts[ii]];
												LOCK(i_buffer_tgts[ii]);
											}
											result = 1;
											if (STRIPLOCK(i_buffer_tgts[ii]) == tgt_block) {
												result = 0;
											}
										}
									}
									if (result == 1) {
										if (tauclosure == TAU_CLOSURE || STRIPLOCK(i_buffer_labels[ii]) != TAU) {
											ISLOCKED(result, i_buffer_labels[ii]);
											if (result == 0) {
												thcont2 = 0;
												signatures_are_equal = 0;
												break;
											}
										}
									}
								}
							}
						}
						// do we need to continue?
						thcont = (thcont2 == 1 && bii+1 < NR_OF_I_BUFFERS) ? 1 : 0;
						bii = (thcont == 1) ? bii+1 : bii;
						BROADCAST_IF_WARPED;
					}
					thcont = thcont2;
					// all i buffers have been checked for a given pivot buffer. check pivot buffer content for full match if we are not checking tau-closure
					if (tauclosure == NO_TAU_CLOSURE) {
						if (thcont == 1) {
							for (pi = 0; pi < pivot_buffer_size; pi++) {
								ISLOCKED(result, pivot_buffer_labels[pi]);
								if (result == 0) {
									thcont = 0;
									signatures_are_equal = 0;
									break;
								}
							}
						}
					}
					// prepare for next iteration
					if (thcont == 1) {
						bpi++;
						thcont = (bpi < NR_OF_PIVOT_BUFFERS) ? 1 : 0;
					}
					BROADCAST_IF_WARPED;
				}
				// do we need to unlock locks of state i?
				GLOBAL_MEM_WRITE_OCCURRED(thcont, write_status);
				BROADCAST_IF_WARPED;
				if (warpcont == 1) {
					for (j = trans_start; j < trans_end; j += STEP_SIZE) {
						ii = d_trans_labels[j];
						d_trans_labels[j] = STRIPLOCK(ii);
					}
				}
				// process outcome
				//thcont = 0;
				//result = 0;
				if (signatures_are_equal != NOT_RELEVANT) {
					if (SIGNATURES_EQUAL) {
						if (tauclosure == TAU_CLOSURE) {
							// we have done tau-closure and
							// the signatures were equal, so we conclude (for now) that this tau-transition
							// is invisible. Propagate the corresponding block information
							//if (STRIPLOCK(block) != STRIPLOCK(candidates_buffer[ci])) {
								//PRINTTHREAD(STRIPLOCK(block), candidates_buffer[ci]);
								//PRINTTHREAD(STRIPLOCK(block), ci);
							//	block = candidates_buffer[ci];
							//	SETBLOCKPROPAGATED(block);
								// another tau-closure iteration is required
								//PRINTTHREAD(0, 0);
							//	(*d_cont) = 1;
							//	thcont = 1;
								// do we need to set a visible tau transition to invisible?
								//ISLOCKED(result, candidates_buffer[ci]);
							if (conf_block == EMPTY_CONFBLOCK) {
								conf_block = tgt_block;
							}
							else if (conf_block != EMPTY_CONFBLOCK && conf_block != NA_CONFBLOCK && conf_block != tgt_block) {
								// multiple candidates means that no confluent tau-transition can be present
								//PRINTTHREAD(1, candidates_label_buffer[tgt_block_id]);
								conf_block = NA_CONFBLOCK;
								thcont = 0;
								signatures_are_equal = NOT_RELEVANT;
							}
						}
						else {
							conf_block = tgt_block;
						}
					}
				}
				// Check for next pivot candidate in buffer
				thcont = (++ci < candidates_buffer_size) ? 1 : 0;
				BROADCAST_IF_WARPED;
			}
			// (if the buffer has overflown, we need to lock candidates in global memory)
			//pi = thcont;
			//ISLOCKED_CANDIDATES(thcont, write_status);
			//BROADCAST_IF_WARPED;
			//if (warpcont == 1) {
			//	for (ci = 0; ci < candidates_buffer_size; ci++) {
			//		for 
			//	}
			//}
			
			// if the candidates buffer has overflown, we need another iteration of while-loop
			ISLOCKED_CANDIDATES(thcont, write_status);
			UNLOCK_CANDIDATES(write_status);
			BROADCAST_IF_WARPED;
		} // end of while-loop over candidate buffers
		// write results to global memory
		// && conf_block != NA_CONFBLOCK
		thcont = (conf_block != EMPTY_CONFBLOCK) ? 1 : 0;
		BROADCAST_IF_WARPED;
		if (warpcont == 1) {
			// closure computation: invisible tau-transition detected
			if (tauclosure == TAU_CLOSURE) {
				if (thcont == 1) {
					//printf("0: Seen by thread %d: %d, [(%d,%d),(%d,%d),(%d,%d),(%d,%d),(%d,%d),(%d,%d)]\n", (blockIdx.x*blockDim.x)+threadIdx.x, conf_block, HASLOCK(closure_buffer[0]),STRIPLOCK(closure_buffer[0]),HASLOCK(closure_buffer[1]),STRIPLOCK(closure_buffer[1]),HASLOCK(closure_buffer[2]),STRIPLOCK(closure_buffer[2]),HASLOCK(closure_buffer[3]),STRIPLOCK(closure_buffer[3]),HASLOCK(closure_buffer[4]),STRIPLOCK(closure_buffer[4]),HASLOCK(closure_buffer[5]),STRIPLOCK(closure_buffer[5]));
					// if state i is a pivot, indicate globally that states need to be moved to the new block later
					//if (STRIPLOCK(block) == i) {
						//(*d_cont2) = 1;
					//}
					if (conf_block == NA_CONFBLOCK) {
						block = NOBLOCK;
						LOCK(block);						
					}
					else if (conf_block != STRIPLOCK(block)){
						block = conf_block;
						LOCK(block);
						(*d_cont) = 1;
					}
				}
				d_results[i] = block;
			}
		}
		//if (conf_block == NA_CONFBLOCK) {
		//	printf("1: Seen by thread %d: [(%d,%d),(%d,%d),(%d,%d),(%d,%d),(%d,%d),(%d,%d)]\n", (blockIdx.x*blockDim.x)+threadIdx.x, HASLOCK(closure_buffer[0]),STRIPLOCK(closure_buffer[0]),HASLOCK(closure_buffer[1]),STRIPLOCK(closure_buffer[1]),HASLOCK(closure_buffer[2]),STRIPLOCK(closure_buffer[2]),HASLOCK(closure_buffer[3]),STRIPLOCK(closure_buffer[3]),HASLOCK(closure_buffer[4]),STRIPLOCK(closure_buffer[4]),HASLOCK(closure_buffer[5]),STRIPLOCK(closure_buffer[5]));
		//}
		if (thcont == 0) {
			thcont = (conf_block == EMPTY_CONFBLOCK && signatures_are_equal != NOT_RELEVANT) ? 1 : 0;
		}
		else {
			thcont = 0;
		}
		BROADCAST_IF_WARPED;
		if (warpcont == 1) {
			// strong comparison: state needs to move to a new block
			if (tauclosure != TAU_CLOSURE) {
				if (thcont == 1) {
					LOCK(trans_start);
					(*d_cont) = 1;
					d_offsets[i] = trans_start;
				}
			}
			//else {
				// make previously 'confluent' state pivot for new block
			//	if (thcont == 1) {
			//		//block = EMPTYTRANS;
			//		LOCK(trans_start);
			//	}
			//	d_results[i] = block;
			//}
		}
	} // end of for-loop over entries i
}
		// write new block info
		//d_results[i] = block;
		//}
	// setting visible tau to invisible
//				thcont = result;
//				BROADCAST_IF_WARPED;
//				if (warpcont == 1) {
//					for (j = trans_start; j < trans_end && warpcont == 1; j += STEP_SIZE) {
//						bpi = d_trans_labels[j];
//						pi = d_trans_tgts[j];
//						if (thcont == 1) {
//							thcont = 0;
//							if (bpi == VISIBLE_TAU) {
//								pi = d_results[pi];
//								if (STRIPLOCK(pi) == STRIPLOCK(candidates_buffer[ci])) {
//									bpi = TAU;
//									thcont = 1;
//								}
//							}
//						}
//						BROADCAST_IF_WARPED;
//						if (warpcont == 1) {
//							d_trans_labels[j] = bpi;
//						}
//						thcont = (STRIPLOCKS(bpi) == TAU) ? 1 : 0;
//						BROADCAST_IF_WARPED;
//					}
//				}
//	thcont = 0;
//	result = 0;
//	if (signatures_are_equal == 0) {
//		// in strong comparisons, make state i candidate for next pivot selection, since no matching pivot has been found
//		if (tauclosure == NO_TAU_CLOSURE) {
//			thcont = 1;
//			LOCK(trans_start);
//			(*d_cont) = 1;
//		}
//	}
//	BROADCAST
//	if (warpcont == 1) {
//		// write new offsets info
//		d_offsets[i] = trans_start;
//	}
		// if there are no more invisible taus, state no longer needs closure checking
//		thcont = 0;
//		if (candidates_buffer_size == 0) {
//			NEEDSCLOSURECHECK(result, block);
//			if (result == 1) {
//				REMOVECLOSURECHECK(block);
//				thcont = 1;
//			}
//		}
//		BROADCAST;
//		if (warpcont == 1) {
//			d_results[i] = block;
//		}


/* // cleanup d_trans_labels locks after signature comparisons
__global__ void postprocess_compare_signatures(uint32_t *d_trans_labels) {
	uint32_t current, result, thcont, warpcont;
	
	for (i = GLOBAL_THREADID; i < d_ntrans; i += NROFTHREADS) {
		thcont = 0;
		current = d_trans_labels[i];
		ISLOCKED(result, current);
		if (result == 1) {
			thcont = 1;
			UNLOCK(current);
		}
		// if whole warp is active in for-loop, coalesce writing
		if (GLOBAL_LASTINWARP_TID < d_ntrans) {
			BROADCAST;
			if (warpcont == 1) {
				d_trans_labels[i] = current;
			}
		}
		else {
			d_trans_labels[i] = current;
		}
	}
} */

//// function to set all values in d_results to an initial value
//__global__ void init_results(uint32_t *d_results) {
//	for (int i = GLOBAL_THREADID; i < d_nstates; i += NROFTHREADS) {
//		d_results[i] = 0;
//	}
//}

/**
 * Host function that reads AUT file (LTS), prepares data for GPU, and launches reduction procedure.
 */
int main(int argc, char** argv) {
	int nblocks = 0;
	int nblocks_trans = 1;
	int nthreadsperblock = BLOCK_SIZE;
	// level of verbosity (1=print progress)
	int verbosity = 0;
	// clock to measure time
	clock_t start, stop;
	double runtime = 0.0;
	string fn;
	uint32_t i, j, k, l;

	// Structures to store LTS input
	uint32_t *offsets, *segments_final, *trans_labels, *trans_tgts, *trans_final_labels, *trans_final_tgts, *statemap, *rev_statemap, trans_cnt, segments_cnt, trans_final_cnt;
	//uint32_t offsets_cnt;
	uint32_t trans_segmentinterval = 0;
	uint32_t dynamic_segments = 0;
	uint32_t states_endofsegments = 0;
	uint32_t nstates, ntrans, nlabels, nsegments, taus_present;
	uint32_t cont, cont2;
	uint32_t branching = 0;
	uint32_t overview = 0;
	uint32_t *overview_array;
	uint32_t thread_buffer_size = 10;
	uint32_t device = 0;

	cudaDeviceProp prop;
	int nDevices;

	// GPU side versions of the input
	//uint32_t *d_offsets, *d_trans_labels, *d_trans_tgts, *d_results;
	uint32_t *d_data;

	if (argc == 1) {
		fprintf(stderr, "ERROR: No input AUT file given!\n");
		exit(1);
	}
	fn = string(argv[1]);

	i = 2;
	while (i < argc) {
		if (!strcmp(argv[i],"-b")) {
			// store nr of blocks to be used
			nblocks = atoi(argv[i+1]);
			i += 2;
		}
		else if (!strcmp(argv[i],"-t")) {
			// store nr of threads per block to be used
			nthreadsperblock = atoi(argv[i+1]);
			i += 2;
		}
		else if (!strcmp(argv[i],"-v")) {
			// store verbosity level
			verbosity = atoi(argv[i+1]);
			if (verbosity > 3) {
				verbosity = 3;
			}
			i += 2;
		}
		else if (!strcmp(argv[i],"-p")) {
			// create 'interleaved segments intervals' in transitions arrays of given size
			trans_segmentinterval = atoi(argv[i+1]);
			i += 2;
		}
		else if (!strcmp(argv[i],"-pd")) {
			// create 'interleaved segments intervals' in transitions arrays dynamically, using the average branching factor as bound
			dynamic_segments = 1;
			i++;
		}
		else if (!strcmp(argv[i],"-br")) {
			// perform branching instead of strong bisimulation reduction
			branching = 1;
			i++;
		}
		else if (!strcmp(argv[i],"-a")) {
			// give an overview of the nr. of outgoing transitions per state
			overview = 1;
			i++;
		}
		else if (!strcmp(argv[i],"-d")) {
			// choose device
			device = atoi(argv[i+1]);
			i += 2;
		}
	}
	fprintf (stdout, "type of bisimulation reduction to be performed: ");
	if (branching == 1) {
		fprintf (stdout, "branching\n");
	}
	else {
		fprintf (stdout, "strong\n");
	}

	ifstream fp (fn.c_str()); 

	// read file header
	string s;
	getline(fp, s);
	// strip 'des(', ')'
	s = s.substr(4).substr(0,s.size()-5);

	istringstream ss(s);
	vector <string> data;
	
	while (ss) {
		string s;
		if (!getline (ss, s, ',')) break;
		if (s[0] == ' ') {
			s = s.substr(1);
		}
		data.push_back(s);
	}
	// process header
	nstates = (uint32_t) atoi(data[2].c_str());
	fprintf(stdout, "nr of states in LTS: %d\n", nstates);
	ntrans = (uint32_t) atoi(data[1].c_str());
	fprintf(stdout, "nr of transitions in LTS: %d\n", ntrans);
	trans_cnt = ntrans;

	// current offset index
	i = 0;
	// current trans index
	j = 0;
	// last source
	k = 0;
	// fill the arrays with the LTS transitions
	offsets = (uint32_t*) malloc(sizeof(uint32_t)*(nstates+1));
	trans_labels = (uint32_t*) malloc(sizeof(uint32_t)*ntrans);
	trans_tgts = (uint32_t*) malloc(sizeof(uint32_t)*ntrans);
	offsets[i++] = 0;

	// keep track of highest transition label encountered (assumes that LTS has numbers as labels!)
	nlabels = 0;
	taus_present = 0;
	while (fp) {
		// Read the input
		string s;
		if (!getline(fp, s)) break;
		
		// strip '(', ')'
		s = s.substr(1).substr(0,s.size()-2);

		istringstream ss(s);
		vector <string> data;
		
		while (ss) {
			string s;
			if (!getline (ss, s, ',')) break;
			if (s[0] == ' ') {
				s = s.substr(1);
			}
			data.push_back(s);
		}
		while (atoi(data[0].c_str()) > k) {
			offsets[i++] = j;
			k++;
		}
		// store label id
		// is label tau?
		if (data[1] == "i") {
			trans_labels[j] = 0;
			taus_present = 1;
			// trick to make strong bisimulation checking work when taus are present:
			// alter the tau-labels to make them visible in the kernel
			if (branching == 0) {
				SETVISIBLE(trans_labels[j]);
			}
		}
		else {
			// strip '"', '"'
			//data[1] = data[1].substr(1).substr(0,data[1].size()-2);
			trans_labels[j] = atoi(data[1].c_str());
			nlabels = MAX(nlabels, trans_labels[j]);
		}
		// store target
		trans_tgts[j++] = atoi(data[2].c_str());
		//cout << k << " -" << trans_labels[j-1] << "-> " << trans_tgts[j-1] << endl;
	}
	// finalise offsets (take care of deadlock states at the end)
	while (i < nstates+1) {
		offsets[i++] = j;
	}
	// if taus are present, increment number of labels
	if (taus_present == 1) {
		nlabels++;
	}
	fprintf(stdout, "nr of transition labels in LTS: %d\n", nlabels);

	// Do we need to give an overview of the nr. of outgoing transitions per state?
	if (overview) {
		// determine max. nr. of outgoing transitions
		j = 0;
		for (i = 0; i < nstates; i++) {
			if (j < (offsets[i+1]-offsets[i])) {
				j = offsets[i+1]-offsets[i];
			}
		}
		// allocate enough memory for analysis
		overview_array = (uint32_t*) malloc(sizeof(uint32_t) * j);
		for (i = 0; i < nstates; i++) {
			overview_array[offsets[i+1]-offsets[i]-1]++;
		}
		fprintf(stdout, "Overview of nr. of outgoing transitions per state:\n");
		for (i = 0; i < j; i++) {
			if (overview_array[i] > 0) {
				fprintf(stdout, "nr. of states with %d successors: %d\n", i+1, overview_array[i]);
			}
		}
		free(overview_array);
		exit(0);
	}

	// Query the device properties and determine max number of integers that we can work with
	cudaSetDevice(device);
	cudaGetDeviceCount(&nDevices);
	if (nDevices == 0) {
		fprintf (stderr, "ERROR: No CUDA compatible GPU detected!\n");
		exit(1);
	}
	cudaGetDeviceProperties(&prop, device);
	fprintf (stdout, "global mem: %lu\n", (size_t) prop.totalGlobalMem);
	fprintf (stdout, "shared mem per block: %d\n", (int) prop.sharedMemPerBlock);
	fprintf (stdout, "max. threads per block: %d\n", (int) prop.maxThreadsPerBlock);
	fprintf (stdout, "max. grid size: %lu\n", (size_t) prop.maxGridSize[0]);
	fprintf (stdout, "nr. of multiprocessors: %d\n", (int) prop.multiProcessorCount);
	
	// choose a suitable size of the thread buffers (depends on nr. of registers available per block)
	thread_buffer_size = (uint32_t) ((((double) prop.regsPerBlock / (double) BLOCK_SIZE) - 28) / 5);
	fprintf (stdout, "suggested thread buffer size: %d\n", thread_buffer_size);

	size_t available, total;
	cudaMemGetInfo(&available, &total);
	
	size_t q_size = 0;
	if (q_size == 0) {
		q_size = total / sizeof(uint32_t);
	}
	size_t el_per_Mb = Mb / sizeof(uint32_t);
	
	while(cudaMalloc((void**)&d_data,  q_size * sizeof(uint32_t)) == cudaErrorMemoryAllocation)	{			
		q_size -= el_per_Mb;
		if( q_size  < el_per_Mb) {
			// signal no free memory
			break;
		}
	}
	// now that we know q_size, free memory again
	cudaFree(d_data);
	fprintf (stdout, "max. nr. of integers that can be stored on GPU: %lu\n", q_size);
	// compute amount of memory required
	j = 2*nstates+2*trans_cnt+5;
	if (j > q_size) {
		fprintf (stderr, "not enough memory on GPU available!\n");
		exit(-1);
	}
	// if dynamic segments are requested, set trans_segmentinterval to av. branching factor
	if (dynamic_segments == 1) {
		trans_segmentinterval = (uint32_t) (ntrans / nstates);
		fprintf (stdout, "segment bound set at %d\n", trans_segmentinterval);
	}
	// Restructure trans array if segment interval > 1
	if (trans_segmentinterval > 1) {
		// first reorder states according to outgoing transitions
		statemap = (uint32_t*) malloc(sizeof(uint32_t)*nstates);
		j = 0;
		for (i = 0; i < nstates; i++) {
			if ((offsets[i+1]-offsets[i]) <= trans_segmentinterval) {
				statemap[j++] = i;
			}
			//else {
			//	fprintf (stdout, "%d is excluded from segmentation. Nr. of outgoing transitions: %d\n", i, offsets[i+1] - offsets[i]);
			//}
		}
		// indicate start of 'other states' section
		// we make sure that this start is a multiple of WARP_SIZE (only whole segments in first part)
		states_endofsegments = ((uint32_t) (j / WARP_SIZE)) * WARP_SIZE;
		fprintf (stdout, "Nr. of states placed in transition segments: %d\n", states_endofsegments);
		for (i = 0; i < nstates; i++) {
			if ((offsets[i+1]-offsets[i]) > trans_segmentinterval) {
				statemap[j++] = i;
			}
		}
		// compute reverse of statemap
		rev_statemap = (uint32_t*) malloc(sizeof(uint32_t)*nstates);
		for (i = 0; i < nstates; i++) {
			rev_statemap[statemap[i]] = i;
		}
		// perform restructuring according to statemap
		// calculate nr. of padding entries required
		j = 0;
		for (i = 0; i < states_endofsegments; i++) {
			j += trans_segmentinterval - (offsets[statemap[i]+1]-offsets[statemap[i]]);
		}
		fprintf (stdout, "Nr. of padding transition entries required: %d\n", j);
		// create final trans and offsets arrays
		trans_final_labels = (uint32_t*) malloc(sizeof(uint32_t) * (ntrans+j));
		trans_final_tgts   = (uint32_t*) malloc(sizeof(uint32_t) * (ntrans+j));
		// add transitions interleaved in segments of WARP_SIZE states (= one warp)
		nsegments = (uint32_t) states_endofsegments/WARP_SIZE;
		segments_final = (uint32_t*) malloc(sizeof(uint32_t) * (nstates+1));
		segments_cnt = 0; trans_final_cnt = 0;
		for (l = 0; l < nsegments; l++) {
			for (i = 0; i < WARP_SIZE; i++) {
				segments_final[segments_cnt++] = trans_final_cnt+i;
			}
			for (i = 0; i < trans_segmentinterval; i++) {
				for (j = 0; j < WARP_SIZE; j++) {
					if (offsets[statemap[l*WARP_SIZE+j]+1] - offsets[statemap[l*WARP_SIZE+j]] <= i) {
						// add padding
						trans_final_labels[trans_final_cnt] = EMPTYTRANS;
						trans_final_tgts[trans_final_cnt++] = EMPTYTRANS;
					}
					else {
						trans_final_labels[trans_final_cnt] = trans_labels[offsets[statemap[l*WARP_SIZE+j]]+i];
						trans_final_tgts[trans_final_cnt++] = trans_tgts[offsets[statemap[l*WARP_SIZE+j]]+i];
					}
				}
			}
		}
		// add final segment border
		segments_final[segments_cnt++] = trans_final_cnt;
		// add the remaining transition entries as they were originally (without interleaving)
		for (i = states_endofsegments; i < nstates; i++) {
			for (j = 0; j < (offsets[statemap[i]+1]-offsets[statemap[i]]); j++) {
				trans_final_labels[trans_final_cnt] = trans_labels[offsets[statemap[i]]+j];
				trans_final_tgts[trans_final_cnt++] = trans_tgts[offsets[statemap[i]]+j];
			}
			segments_final[segments_cnt++] = trans_final_cnt;
		}
		// At this stage, check if the new structures fit in the GPU global memory. If not,
		// revert to the original ones
		j = 2*nstates+2*trans_final_cnt+5;
		if (j > q_size) {
				fprintf (stdout, "restructured graph does not fit in the GPU memory. Reverting back to original one.\n");
				free(segments_final);
				free(trans_final_labels);
				free(trans_final_tgts);
				trans_segmentinterval = 0;
				states_endofsegments = 0;
		}
		else {
			// Now rewrite all the trans entries in trans_final, to reflect statemap
			for (i = 0; i < trans_final_cnt; i++) {
				if (trans_final_labels[i] != EMPTYTRANS) {
					trans_final_tgts[i] = rev_statemap[trans_final_tgts[i]];
				}
			}
			// make the new structures the default ones
			free(offsets);
			free(trans_labels);
			free(trans_tgts);
			//offsets_cnt = segments_cnt;
			trans_cnt = trans_final_cnt;
			offsets = segments_final;
			trans_labels = trans_final_labels;
			trans_tgts = trans_final_tgts;
		}
		free(statemap);
		free(rev_statemap);
	}

	// determine actual nr of threads per block
	nthreadsperblock = MAX(1,MIN(prop.maxThreadsPerBlock,nthreadsperblock));
	// determine actual nr of blocks
	if (nblocks == 0) {
		// round up
		nblocks = MAX(1,(nstates + (nthreadsperblock-1))/ nthreadsperblock);
		nblocks_trans = MAX(1,(ntrans + (nthreadsperblock-1))/ nthreadsperblock);
	}
	if (nblocks > 0) {
		nblocks = MAX(1,MIN(prop.maxGridSize[0],nblocks));
		nblocks_trans = nblocks;
	}
	else {
		// set to maximum nr
		nblocks = prop.maxGridSize[0];
		nblocks_trans = nblocks;
	}
	fprintf (stdout, "%d blocks will be launched on the GPU\n", nblocks);
	fprintf (stdout, "%d threads will be launched per block\n", nthreadsperblock);
	
	fprintf (stdout, "launching reduction\n");
	// Start timer
	assert((start = clock())!=-1);
	// Allocate memory on GPU
	cudaMalloc((void**)&d_data,  j * sizeof(uint32_t));
	// Copy data to GPU
	CUDA_CHECK_RETURN(cudaMemcpy(&(d_data[s_offsets]), offsets, (nstates+1)*sizeof(uint32_t), cudaMemcpyHostToDevice))
	CUDA_CHECK_RETURN(cudaMemcpy(&(d_data[s_trans_labels]), trans_labels, trans_cnt*sizeof(uint32_t), cudaMemcpyHostToDevice))
	CUDA_CHECK_RETURN(cudaMemcpy(&(d_data[s_trans_tgts]), trans_tgts, trans_cnt*sizeof(uint32_t), cudaMemcpyHostToDevice))
	// copy symbols
	cudaMemcpyToSymbol(d_states_endofsegments, &states_endofsegments, sizeof(uint32_t));
	cudaMemcpyToSymbol(d_nstates, &nstates, sizeof(uint32_t));
	cudaMemcpyToSymbol(d_ntrans, &trans_cnt, sizeof(uint32_t));
	cudaMemcpyToSymbol(d_nlabels, &nlabels, sizeof(uint32_t));
	cudaMemcpyToSymbol(d_trans_segmentinterval, &trans_segmentinterval, sizeof(uint32_t));

	// free LTS from host memory
	free(offsets);
	free(trans_labels);
	free(trans_tgts);

	init_results<<<nblocks, nthreadsperblock>>>(d_data);

	uint32_t zero = 0;

	// compute tau-closures
//	if (branching == 1 && taus_present == 1) {
//		cont = 1;
//		CUDA_CHECK_RETURN(cudaMemcpy(&(d_data[s_cont2]), &zero, sizeof(uint32_t), cudaMemcpyHostToDevice));
//		while (cont == 1) {
//			CUDA_CHECK_RETURN(cudaMemcpy(&(d_data[s_cont]), &zero, sizeof(uint32_t), cudaMemcpyHostToDevice));
//			//fprintf (stdout, "closure checking\n");
//			compare_signatures<<<nblocks, nthreadsperblock, nthreadsperblock * sizeof(uint32_t)>>>(&(d_data[s_offsets]), &(d_data[s_trans_labels]), &(d_data[s_trans_tgts]), &(d_data[s_results]), &(d_data[s_cont]), &(d_data[s_cont2]), TAU_CLOSURE);
//			//print_signatures((&d_data[s_offsets]), (&d_data[s_trans_labels]), (&d_data[s_trans_tgts]), (&d_data[s_results]), nstates, ntrans);
//			//print_outputsize(&(d_data[s_results]), nstates);		
//			//print_status(&(d_data[s_offsets]), &(d_data[s_trans_labels]), &(d_data[s_trans_tgts]), &(d_data[s_results]), nstates, trans_cnt);
//			// are there still states that require closure checking?
//			CUDA_CHECK_RETURN(cudaMemcpy(&cont, &(d_data[s_cont]), sizeof(uint32_t), cudaMemcpyDeviceToHost));
//		}
//		CUDA_CHECK_RETURN(cudaMemcpy(&cont2, &(d_data[s_cont2]), sizeof(uint32_t), cudaMemcpyDeviceToHost));
//		// possibly move states to new block if a pivot moved
//		if (cont2 == 1) {
//			postprocess_tauclosures<<<nblocks, nthreadsperblock, nthreadsperblock * sizeof(uint32_t)>>>(&(d_data[s_results]));			
//		}
//		// detect visible tau-transitions
//		//fprintf (stdout, "detecting visible tau-transitions\n");
//		detect_visible_taus<<<nblocks, nthreadsperblock, nthreadsperblock * sizeof(uint32_t)>>>(&(d_data[s_offsets]), &(d_data[s_trans_labels]), &(d_data[s_trans_tgts]), &(d_data[s_results]));
//
//		//print_invisiblestates(&(d_data[s_results]), nstates);
//
//		//if (cont == 1) {
//			//fprintf (stdout, "NEW VISIBLE TAUS DETECTED!\n");
//		//	cont_visible = 1;
//		//}
//		//print_signatures((&d_data[s_offsets]), (&d_data[s_trans_labels]), (&d_data[s_trans_tgts]), (&d_data[s_results]), nstates, ntrans);
//		//print_outputsize(&(d_data[s_results]), nstates);
//		
//		// select candidates for pivots
//		identify_candidates_bb<<<nblocks, nthreadsperblock, nthreadsperblock * sizeof(uint32_t)>>>(&(d_data[s_offsets]), &(d_data[s_results]));	
//	}
//	else {
		// select candidates for pivots
	identify_candidates_sb<<<nblocks, nthreadsperblock>>>(&(d_data[s_offsets]));
//	}
//	cont = 1;
//	j = 0; k = 0;
//	while (cont == 1) {
//		i = 0;
//		while (cont == 1) {
//			fprintf (stdout, "pivot selection\n");
//			CUDA_CHECK_RETURN(cudaMemcpy(&(d_data[s_cont]), &zero, sizeof(uint32_t), cudaMemcpyHostToDevice));
//			CUDA_CHECK_RETURN(cudaMemcpy(&(d_data[s_cont2]), &zero, sizeof(uint32_t), cudaMemcpyHostToDevice));
//			choose_representatives<<<nblocks, nthreadsperblock, nthreadsperblock * sizeof(uint32_t)>>>(&(d_data[s_offsets]), &(d_data[s_trans_labels]), &(d_data[s_trans_tgts]), &(d_data[s_results]), &(d_data[s_cont]), &(d_data[s_cont2]), j, i);
//			CUDA_CHECK_RETURN(cudaMemcpy(&cont2, &(d_data[s_cont2]), sizeof(uint32_t), cudaMemcpyDeviceToHost));
//			if (cont2 == 1) {
//				postprocess_representatives<<<nblocks_trans, nthreadsperblock>>>(&(d_data[s_offsets]), &(d_data[s_trans_tgts]), &(d_data[s_results]));
//			}
//			CUDA_CHECK_RETURN(cudaMemcpy(&cont, &(d_data[s_cont]), sizeof(uint32_t), cudaMemcpyDeviceToHost));
//			i++;
//		}
//		fprintf(stdout, "strong\n");
//		compare_signatures<<<nblocks, nthreadsperblock, nthreadsperblock * sizeof(uint32_t)>>>(&(d_data[s_offsets]), &(d_data[s_trans_labels]), &(d_data[s_trans_tgts]), &(d_data[s_results]), &(d_data[s_cont]), &(d_data[s_cont2]), 2);		
//		print_outputsize(&(d_data[s_results]), nstates);
//		//print_signatures((&d_data[s_offsets]), (&d_data[s_trans_labels]), (&d_data[s_trans_tgts]), (&d_data[s_results]), nstates, ntrans);
//		CUDA_CHECK_RETURN(cudaMemcpy(&cont, &(d_data[s_cont]), sizeof(uint32_t), cudaMemcpyDeviceToHost));
//		if (k == 0) {
//			j++;
//			if (j == nlabels) {
//				j = 0;
//				k = 1;
//			}
//		}
//	}
//	print_outputsize(&(d_data[s_results]), nstates);
	
	cont = 1;
	// perform strong comparisons, until fix-point has been reached
	// j is used to determine which outgoing transition label to consider when splitting blocks (jth)
	// k is used to keep track of whether the maximum value for j (nlabels) has been reached once
	j = 0; k = 0;
	//fprintf (stdout, "strong checking\n");
	//cont_visible = 0;
	//uint32_t counter = 0;
	while (cont == 1) {
		// start representatives selection procedure
		i = 0;
		if (verbosity > 2) {
			fprintf (stdout, "pivot selection\n");
		}
		while (cont == 1) {
			CUDA_CHECK_RETURN(cudaMemcpy(&(d_data[s_cont]), &zero, sizeof(uint32_t), cudaMemcpyHostToDevice));
			CUDA_CHECK_RETURN(cudaMemcpy(&(d_data[s_cont2]), &zero, sizeof(uint32_t), cudaMemcpyHostToDevice));
			choose_representatives<<<nblocks, nthreadsperblock, nthreadsperblock * sizeof(uint32_t)>>>(&(d_data[s_offsets]), &(d_data[s_trans_labels]), &(d_data[s_trans_tgts]), &(d_data[s_results]), &(d_data[s_cont]), &(d_data[s_cont2]), j, i);
			CUDA_CHECK_RETURN(cudaMemcpy(&cont2, &(d_data[s_cont2]), sizeof(uint32_t), cudaMemcpyDeviceToHost));
			if (cont2 == 1) {
				postprocess_representatives<<<nblocks_trans, nthreadsperblock>>>(&(d_data[s_offsets]), &(d_data[s_trans_tgts]), &(d_data[s_results]));
			}
			CUDA_CHECK_RETURN(cudaMemcpy(&cont, &(d_data[s_cont]), sizeof(uint32_t), cudaMemcpyDeviceToHost));
			i++;
		}
		//print_outputsize(&(d_data[s_results]), nstates);

		if (verbosity > 2) {
			print_signatures((&d_data[s_offsets]), (&d_data[s_trans_labels]), (&d_data[s_trans_tgts]), (&d_data[s_results]), nstates, ntrans);
			print_outputsize(&(d_data[s_results]), nstates);
		}
			//init_results<<<nblocks, nthreadsperblock>>>(&(d_data[s_results]));
			//identify_candidates_sb<<<nblocks, nthreadsperblock>>>(&(d_data[s_offsets]));

		// compute tau-closures
		if (branching == 1 && taus_present == 1) {
			cont = 1;
			CUDA_CHECK_RETURN(cudaMemcpy(&(d_data[s_cont2]), &zero, sizeof(uint32_t), cudaMemcpyHostToDevice));
			if (verbosity > 2) {
				fprintf (stdout, "closure checking\n");
			}
			while (cont == 1) {
				CUDA_CHECK_RETURN(cudaMemcpy(&(d_data[s_cont]), &zero, sizeof(uint32_t), cudaMemcpyHostToDevice));
				compare_signatures<<<nblocks, nthreadsperblock, nthreadsperblock * sizeof(uint32_t)>>>(&(d_data[s_offsets]), &(d_data[s_trans_labels]), &(d_data[s_trans_tgts]), &(d_data[s_results]), &(d_data[s_cont]), &(d_data[s_cont2]), TAU_CLOSURE);
				//print_signatures((&d_data[s_offsets]), (&d_data[s_trans_labels]), (&d_data[s_trans_tgts]), (&d_data[s_results]), nstates, ntrans);
				//print_outputsize(&(d_data[s_results]), nstates);		
				//print_status(&(d_data[s_offsets]), &(d_data[s_trans_labels]), &(d_data[s_trans_tgts]), &(d_data[s_results]), nstates, trans_cnt);
				// are there still states that require closure checking?
				CUDA_CHECK_RETURN(cudaMemcpy(&cont, &(d_data[s_cont]), sizeof(uint32_t), cudaMemcpyDeviceToHost));
			}
			//CUDA_CHECK_RETURN(cudaMemcpy(&cont2, &(d_data[s_cont2]), sizeof(uint32_t), cudaMemcpyDeviceToHost));
			// possibly move states to new block if a pivot moved
			//if (cont2 == 1) {
			//	fprintf (stdout, "FIX!\n");
			//	postprocess_tauclosures<<<nblocks, nthreadsperblock, nthreadsperblock * sizeof(uint32_t)>>>(&(d_data[s_results]));			
			//}
			// detect visible tau-transitions
			//fprintf (stdout, "detecting visible tau-transitions\n");
			detect_visible_taus<<<nblocks, nthreadsperblock, nthreadsperblock * sizeof(uint32_t)>>>(&(d_data[s_offsets]), &(d_data[s_trans_labels]), &(d_data[s_trans_tgts]), &(d_data[s_results]));
			//cont_visible = cont;
			if (verbosity > 2) {
				print_signatures((&d_data[s_offsets]), (&d_data[s_trans_labels]), (&d_data[s_trans_tgts]), (&d_data[s_results]), nstates, ntrans);
				print_outputsize(&(d_data[s_results]), nstates);
			}
		}
		//print_signatures((&d_data[s_offsets]), (&d_data[s_trans_labels]), (&d_data[s_trans_tgts]), (&d_data[s_results]), nstates, ntrans);
		//print_invisiblestates(&(d_data[s_results]), nstates);

		if (verbosity > 2) {
			fprintf (stdout, "strong checking\n");
		}
		compare_signatures<<<nblocks, nthreadsperblock, nthreadsperblock * sizeof(uint32_t)>>>(&(d_data[s_offsets]), &(d_data[s_trans_labels]), &(d_data[s_trans_tgts]), &(d_data[s_results]), &(d_data[s_cont]), &(d_data[s_cont2]), NO_TAU_CLOSURE);
		if (verbosity > 2) {
			print_signatures((&d_data[s_offsets]), (&d_data[s_trans_labels]), (&d_data[s_trans_tgts]), (&d_data[s_results]), nstates, ntrans);
			print_outputsize(&(d_data[s_results]), nstates);		
		}
		CUDA_CHECK_RETURN(cudaMemcpy(&cont, &(d_data[s_cont]), sizeof(uint32_t), cudaMemcpyDeviceToHost));
		if (k == 0) {
			j++;
			if (j == nlabels) {
				j = 0;
				k = 1;
			}
		}
		//print_signatures((&d_data[s_offsets]), (&d_data[s_trans_labels]), (&d_data[s_trans_tgts]), (&d_data[s_results]), nstates, ntrans);
		//fprintf(stdout, "%d\n", counter++);
		if (verbosity > 1) {
			print_outputsize(&(d_data[s_results]), nstates);
		}
	}
	// determine runtime
	stop = clock();
	runtime = (double) (stop-start)/CLOCKS_PER_SEC;
	fprintf (stdout, "Run time: %f\n", runtime);

	if (verbosity > 0) {
		print_outputsize(&(d_data[s_results]), nstates);
	}
	return 0;
}
