/* *
 * Copyright 1993-2012 NVIDIA Corporation.  All rights reserved.
 *
 * Please refer to the NVIDIA end user license agreement (EULA) associated
 * with this source code for terms and conditions that govern your use of
 * this software. Any use, reproduction, disclosure, or distribution of
 * this software and related documentation outside the terms of the EULA
 * is strictly prohibited.
 */

/* *
 * Read an .AUT file, hide fixed numbers of actions (roughly 25%, 50%, and 75%), perform tau-compression, and save the resulting LTSs, v.1.0 by Anton Wijs, 2014
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <time.h>
#include <math.h>
#include <map>
#include "uthash.h"

static const unsigned int IOBUFFERSIZE = 100;
static const unsigned int EMPTYSTATE = 0x7FFFFFFF;
static const unsigned int LAST = 0xFFFFFFFF;


// BIT MANIPULATION MACROS

#define MIN(a,b) \
   ({ __typeof__ (a) _a = (a); \
       __typeof__ (b) _b = (b); \
     _a < _b ? _a : _b; })

#define MAX(a,b) \
   ({ __typeof__ (a) _a = (a); \
       __typeof__ (b) _b = (b); \
     _a > _b ? _a : _b; })

// --- BIT OPERATIONS TO BE PERFORMED ON NUMBER ARRAY (COMBINES NUMBER INFO AND STACK CONTENT LOOKUP
#define GETNUMBER(t)							(number[(t)] & 0x7FFFFFFF)
#define SETNUMBER(t, a)							{number[(t)] = (number[(t)] & 0x80000000) | (a);}
#define ISONSTACK(t)							(number[(t)] >> 31)
#define SETSTACKBIT(t)							{number[(t)] = number[(t)] | 0x80000000;}
#define UNSETSTACKBIT(t)						{number[(t)] = number[(t)] & 0x7FFFFFFF;}
#define STACKISEMPTY							(stackTop == NULL)

// stack
typedef struct node {
	unsigned int x;
	node *next;
} node;

// callstack node
typedef struct callnode {
	unsigned int x;
	unsigned int pred;
	unsigned int offset;
	callnode *next;
} callnode;

// hashable struct for obtaining the final state IDs of the output (new) LTSs
typedef struct hashnode {
	int tauid; // key given be result of tau-compression
	unsigned int id; // new id for a state
	UT_hash_handle hh; // makes node hashable
} hashnode;

// hashable struct for maintaining a set of states
typedef struct statesetnode {
	int id;
	UT_hash_handle hh;
} statesetnode;

// Global pointer to stack head
node *stackTop = NULL;
node *newstackTop = NULL;
// Global pointer to call stack
callnode *callstackTop = NULL;
// Global variables
unsigned int nstates, ntrans, scc_index;
unsigned int *results, *number, *offsets, *transact, *transtgt;
// Array to represent hiding set
unsigned int *hidden;
// output LTS
unsigned int *new_offsets, *new_transact, *new_transtgt, *new_counters, *permutation, *tmparray1, *tmparray2, new_nstates;
// hashtable to assign new IDs to states in new LTS
hashnode *newIDs;
// hashtable to keep a set of states
statesetnode *stateset;
unsigned int newIDcounter = 0;
// offset to tell the compare function which results to look at
unsigned int compare_offset = 0;
// flag to indicate that changes were made in the first iteration
unsigned int changed = 0;

void push(unsigned int a) {
	static node *newItem;
	
	newItem = (node*) calloc(sizeof(node),1);
	newItem->x = a;
	newItem->next = stackTop;
	stackTop = newItem;
	SETSTACKBIT(a);
}

unsigned int pop() {
	static node *headItem;
	unsigned int result;
	
	if (stackTop != NULL) {
		result = stackTop->x;
		headItem = stackTop;
		stackTop = stackTop->next;
		free(headItem);
		UNSETSTACKBIT(result);
		return result;
	}
	else {
		return EMPTYSTATE;
	}
}

void pushcall(unsigned int a, unsigned int p) {
	static callnode *newItem;
	
	newItem = (callnode*) calloc(sizeof(callnode),1);
	newItem->x = a;
	newItem->pred = p;
	newItem->offset = LAST;
	newItem->next = callstackTop;
	callstackTop = newItem;
}

void popcall() {
	static callnode *headItem;
	
	if (callstackTop != NULL) {
		headItem = callstackTop;
		callstackTop = callstackTop->next;
		free(headItem);
	}
}

unsigned int topoffsetcall() {
	if (callstackTop == NULL) {
		return EMPTYSTATE;
	}
	else {
		if (callstackTop->offset == LAST) {
			return offsets[callstackTop->x];
		}
		else {
			return (callstackTop->offset)+1;
		}
	}
}

// Tarjan SCC function over tau-transitions, using callstack
void strongconnect(unsigned int v) {
	unsigned int i, state, start_trans, current_state;
	callnode *currentnode;
	
	pushcall(v, EMPTYSTATE);
	while (callstackTop != NULL) {
		currentnode = callstackTop;
		current_state = currentnode->x;
		if (GETNUMBER(current_state) == EMPTYSTATE) {
			SETNUMBER(current_state, scc_index);
			results[current_state] = scc_index;
			scc_index++;
			push(current_state);
		}
		// successors of current_state
		start_trans = topoffsetcall();
		for (currentnode->offset = start_trans; currentnode->offset < offsets[current_state+1]; (currentnode->offset)++) {
			if (transact[currentnode->offset] == 0 || hidden[transact[currentnode->offset]] == 1) {
				state = transtgt[currentnode->offset];
				if (GETNUMBER(state) == EMPTYSTATE) {
					// add successor to call stack
					pushcall(state, current_state);
					break;
				}
				else if (ISONSTACK(state)) {
					results[current_state] = MIN(results[current_state], GETNUMBER(state));
					// indicate that a change is going to be made to the graph
					changed = 1;
				}
			}
		}
		if (currentnode->offset == offsets[current_state+1]) {
			// is v a root state?
			if (results[current_state] == GETNUMBER(current_state)) {
				new_nstates++;
				state = pop();
				while (state != current_state && state != EMPTYSTATE) {
					results[state] = results[current_state];
					state = pop();
				}
			}
			// update predecessor
			if (callstackTop->pred != EMPTYSTATE) {
				results[callstackTop->pred] = MIN(results[callstackTop->pred], results[current_state]);
			}
			popcall();
		}
	}
}

// hide actions in LTS
void hideAut(unsigned int amount, unsigned int max_actlabel, unsigned int has_hidden_act) {
	unsigned int amounttohide, nr_of_acts;
	int next;
	
	nr_of_acts = max_actlabel + has_hidden_act;
	// if hidden actions are present, and this is the first hiding effort, do nothing
	if (amount != 1 || has_hidden_act == 0) {
		amounttohide = (nr_of_acts / 4) * amount;
		// hide the even numbered actions first
		next = 0;
		while (amounttohide > 0) {
			next += 2;
			if (next > max_actlabel) {
				break;
			}
			amounttohide--;
			if (hidden[next] == 0) {
				hidden[next] = 1;
			}
		}
		// if there is still hiding to do, hide the odd actions
		next = -1;
		while (amounttohide > 0) {
			next += 2;
			if (next > max_actlabel) {
				break;
			}
			amounttohide--;
			if (hidden[next] == 0) {
				hidden[next] = 1;
			}
		}
	}
}

// compare function to sort the output LTS transitions
int compare (const void * a, const void * b)
{
	return ( new_transact[compare_offset+(*(int*)a)] - new_transact[compare_offset+(*(int*)b)] );
}

// HASHTABLE FUNCTIONS

// clean the hashtable
void clean_hashtable() {
	hashnode *current, *tmp;
	
	HASH_ITER(hh, newIDs, current, tmp) {
		HASH_DEL(newIDs, current);
		free(current);
	}
	newIDcounter = 0;
}

// set a new ID mapping to newIDs
void set_ID(unsigned int tauid) {
	hashnode *s;

	int tauidcast = (int) tauid;
	HASH_FIND_INT(newIDs, &tauidcast, s);
	if (s==NULL) {
		s = (hashnode*) malloc(sizeof(hashnode));
		s->tauid = tauidcast;
		s->id = newIDcounter++;
		HASH_ADD_INT(newIDs, tauid, s);
	}
}

// get an ID
unsigned int get_ID(unsigned int tauid) {
	hashnode *s;
	
	int tauidcast = (int) tauid;
	HASH_FIND_INT(newIDs, &tauidcast, s);
	if (s != NULL) {
		return s->id;
	}
	else {
		return EMPTYSTATE;
	}
}

// clean the stateset
void clean_stateset() {
	statesetnode *current, *tmp;
	
	HASH_ITER(hh, stateset, current, tmp) {
		HASH_DEL(stateset, current);
		free(current);
	}
}

// add a state to the stateset. returns 1 or 0 depending on whether
// an insertion was required
int addstate(unsigned int id) {
	statesetnode *s;
	
	int idcast = (int) id;
	HASH_FIND_INT(stateset, &idcast, s);
	if (s==NULL) {
		s = (statesetnode*) malloc(sizeof(statesetnode));
		s->id = idcast;
		HASH_ADD_INT(stateset, id, s);
		return 1;
	}
	else {
		return 0;
	}
}

/**
 * Host function that reads MDP file (of PRISM) and launches algorithm.
 */
int main(int argc, char** argv) {
	FILE *fp;
	// level of verbosity (1=print progress)
	int verbosity = 0;
	char stmp1[IOBUFFERSIZE], stmp2[IOBUFFERSIZE], fn[50], fn2[55];
	unsigned int i, j, k, size;

	// Structures to store AUT input
	unsigned int offsets_cnt, trans_cnt;
	unsigned int max_brfactor;
	unsigned int src_state, tgt_state;
	unsigned int max_actlabel = 0;
	unsigned int has_hidden_act = 0;
	unsigned int lastact;

	if (argc == 1) {
		fprintf(stderr, "ERROR: No input AUT given!\n");
		exit(1);
	}

	strcpy(fn, argv[1]);

	i = 2;
	while (i < argc) {
		if (!strcmp(argv[i],"-v")) {
			// store verbosity level
			verbosity = atoi(argv[i+1]);
			if (verbosity > 3) {
				verbosity = 3;
			}
			i += 2;
		}
	}

	fp = fopen(fn, "r");
	if (fp) {
		// Read the input
		fgets(stmp1, IOBUFFERSIZE, fp);

		char *sp1, *sp2;
		sp1 = stmp1+8;
		sp2 = strchr(sp1, ',');
		memcpy (stmp2, sp1, sp2-sp1);
		stmp2[sp2-sp1] = '\0';
		ntrans = (unsigned int) atoi(stmp2);
		fprintf(stdout, "nr of transitions in LTS: %d\n", ntrans);
		sp1 = sp2+2;
		sp2 = strchr(sp1, ')');
		memcpy (stmp2, sp1, sp2-sp1);
		stmp2[sp2-sp1] = '\0';
		nstates = (unsigned int) atoi(stmp2);
		fprintf(stdout, "nr of states in LTS: %d\n", nstates);

		// fill the arrays with the LTS transitions
		offsets = (unsigned int*) malloc(sizeof(unsigned int)*(nstates+1));
		transact = (unsigned int*) malloc(sizeof(unsigned int)*ntrans);
		transtgt = (unsigned int*) malloc(sizeof(unsigned int)*ntrans);
		offsets_cnt = 0; trans_cnt = 0;
		offsets[offsets_cnt++] = 0;
		j = 0;
		max_brfactor = 0;
		for (unsigned int x = 0; x < ntrans; x++) {
			fgets(stmp1, IOBUFFERSIZE, fp);
			sp1 = stmp1+1;
			sp2 = strchr(sp1, ',');
			memcpy (stmp2, sp1, sp2-sp1);
			stmp2[sp2-sp1] = '\0';
			src_state = (unsigned int) atoi(stmp2);
			if (j != src_state) {
				// new src state. add offset
				for (k = j; k < src_state; k++) {
					offsets[offsets_cnt++] = trans_cnt;
				}
				j = src_state;
				// update max. branching factor
				max_brfactor = MAX(max_brfactor, (trans_cnt - offsets[offsets_cnt-2]));
			}
			// read action label
			sp1 = sp2+2;
			// ignore quotes if required
			if ((*sp1) == '\"') {
				sp1++;
			}
			sp2 = strchr(sp1, ',');
			// ignore quotes if required
			if ((*(sp2-1)) == '\"') {
				sp2--;
			}
			memcpy (stmp2, sp1, sp2-sp1);
			stmp2[sp2-sp1] = '\0';
			if (stmp2[0] == 'i') {
				stmp2[0] = '0';
				has_hidden_act = 1;
			}
			transact[trans_cnt] = (unsigned int) atoi(stmp2);
			max_actlabel = MAX(max_actlabel, transact[trans_cnt]);
			// read target state
			sp1 = strchr(sp2, ' ');
			sp1++;
			sp2 = strchr(sp1, ')');
			memcpy (stmp2, sp1, sp2-sp1);
			stmp2[sp2-sp1] = '\0';
			tgt_state = (unsigned int) atoi(stmp2);
			transtgt[trans_cnt++] = tgt_state;
		}
		// store final offset to end of trans array
        while (offsets_cnt < nstates+1) {
            offsets[offsets_cnt++] = trans_cnt;
		}
		// update max. branching factor
		max_brfactor = MAX(max_brfactor, (trans_cnt - offsets[offsets_cnt-1]));		
		fclose(fp);
		fprintf(stdout, "max. branching factor: %d\n", max_brfactor);
		// allocate memory for hiding set
		hidden = (unsigned int*) malloc(sizeof(unsigned int)*(max_actlabel+1));
		for (i = 0; i < max_actlabel+1; i++) {
			hidden[i] = 0;
		}

		// perform the hiding and tau-compression for 25%, 50%, and 75% hiding
		for (i = 1; i < 4; i++) {
			fprintf (stdout, "hiding actions\n");
			hideAut(i, max_actlabel, has_hidden_act);
			fprintf (stdout, "performing tau-compression\n");
			results = (unsigned int*) malloc(nstates * sizeof(unsigned int));
			number = (unsigned int*) malloc(nstates * sizeof(unsigned int));
			for (j = 0; j < nstates; j++) {
				results[j] = EMPTYSTATE;
				SETNUMBER(j, EMPTYSTATE);
			}
			scc_index = 0;
			new_nstates = 0;
			for (j = 0; j < nstates; j++) {
				if (GETNUMBER(j) == EMPTYSTATE) {
					strongconnect(j);
				}
			}
			if (i != 1 || changed == 1) {
				fprintf (stdout, "constructing new LTS\n");
				// postprocessing of tau-compression is required to ensure that we have an uninterrupted list
				// of state IDs for the new LTS
				for (j = 0; j < nstates; j++) {
					set_ID(results[j]);
				}
				new_offsets = (unsigned int*) malloc(sizeof(unsigned int)*(new_nstates+1));
				for (j = 0; j < new_nstates+1; j++) {
					new_offsets[j] = 0;
				}
				for (j = 0; j < nstates; j++) {
					k = offsets[j+1] - offsets[j];
					size = get_ID(results[j]);
					new_offsets[size+1] += k;
				}
				// determine absolute offsets
				k = 0;
				for (j = 0; j < new_nstates+1; j++) {
					k += new_offsets[j];
					new_offsets[j] = k;
				}
				// fill new trans arrays using counters
				new_counters = (unsigned int*) malloc(sizeof(unsigned int)*new_nstates);
				new_transact = (unsigned int*) malloc(sizeof(unsigned int)*ntrans);
				new_transtgt = (unsigned int*) malloc(sizeof(unsigned int)*ntrans);
				for (j = 0; j < new_nstates; j++) {
					new_counters[j] = new_offsets[j];
				}
				for (j = 0; j < nstates; j++) {
					for (k = offsets[j]; k < offsets[j+1]; k++) {
						if (hidden[transact[k]] == 1) {
							new_transact[new_counters[get_ID(results[j])]] = 0;
						}
						else {
							new_transact[new_counters[get_ID(results[j])]] = transact[k];
						}
						new_transtgt[new_counters[get_ID(results[j])]] = get_ID(results[transtgt[k]]);
						new_counters[get_ID(results[j])]++;
					}
				}
				// sort the transitions by action label. we use a permutation array for this.
				for (j = 0; j < new_nstates; j++) {
					size = new_offsets[j+1] - new_offsets[j];
					compare_offset = new_offsets[j];
					permutation = (unsigned int*) malloc(sizeof(unsigned int)*size);
					tmparray1 = (unsigned int*) malloc(sizeof(unsigned int)*size);
					tmparray2 = (unsigned int*) malloc(sizeof(unsigned int)*size);
					for (k = 0; k < size; k++) {
						permutation[k] = k;
					}
					// do the sorting
					qsort (permutation, size, sizeof(unsigned int), compare);
					// reorder based on final permutation
					for (k = 0; k < size; k++) {
						tmparray1[k] = new_transact[new_offsets[j]+permutation[k]];
						tmparray2[k] = new_transtgt[new_offsets[j]+permutation[k]];
					}
					for (k = 0; k < size; k++) {
						new_transact[new_offsets[j]+k] = tmparray1[k];
						new_transtgt[new_offsets[j]+k] = tmparray2[k];
					}
					free(permutation);
					free(tmparray1);
					free(tmparray2);
				}
				fprintf (stdout, "writing new .AUT file\n");
				strcpy(fn2, fn);
				sp1 = strrchr(fn2, '.');
				(*sp1) = '\0';
				strcat(fn2, "_hx.aut");
				switch (i) {
					case 1:
						(*(sp1+2)) = '1';
						break;
					case 2:
						(*(sp1+2)) = '2';
						break;
					case 3:
						(*(sp1+2)) = '3';
						break;
				}
				// determine nr. of transitions in new LTS before writing
				// size variable is used to keep track of nr of transitions in new LTS
				size = 0;
				for (j = 0; j < new_nstates; j++) {
					lastact = 0;
					clean_stateset();
					for (k = new_offsets[j]; k < new_offsets[j+1]; k++) {
						if (new_transact[k] == 0) {
							if (j != new_transtgt[k]) {
								if (addstate(new_transtgt[k]) == 1) {
									size++;
								}
							}
						}
						else {
							if (new_transact[k] != lastact) {
								lastact = new_transact[k];
								clean_stateset();
							}
							if (addstate(new_transtgt[k]) == 1) {
								size++;
							}
						}
					}
				}
				fp = fopen (fn2, "w");
				fprintf (fp, "des (0, %d, %d)\n", size, new_nstates);
				for (j = 0; j < new_nstates; j++) {
					lastact = 0;
					clean_stateset();
					for (k = new_offsets[j]; k < new_offsets[j+1]; k++) {
						if (new_transact[k] == 0) {
							// ignore tau-selfloops
							if (j != new_transtgt[k]) {
								// is the transition not a duplicate of the previous one?
								if (addstate(new_transtgt[k]) == 1) {
									fprintf (fp, "(%d, i, %d)\n", j, new_transtgt[k]);
								}
							}
						}
						else {
							// is the transition not a duplicate of the previous one?
							if (new_transact[k] != lastact) {
								lastact = new_transact[k];
								clean_stateset();
							}
							if (addstate(new_transtgt[k]) == 1) {
								fprintf (fp, "(%d, %d, %d)\n", j, new_transact[k], new_transtgt[k]);						
								lastact = new_transact[k];
							}
						}
					}
				}
				clean_stateset();
				fclose(fp);
				free(new_counters);
				free(transact);
				free(transtgt);
				free(offsets);
				// replace the old LTS with the new one
				offsets = new_offsets;
				transact = new_transact;
				transtgt = new_transtgt;
				nstates = new_nstates;
				ntrans = new_offsets[new_nstates];
			}
			else {
				fprintf (stdout, "LTS remains as is\n");
			}
			free(results);
            free(number);
			clean_hashtable();
			fprintf (stdout, "finished!\n");
		}
        free(offsets);
        free(transact);
        free(transtgt);
        free(hidden);
	}
	else {
		fprintf(stderr, "ERROR: input file does not exist!\n");
		exit(1);
	}
}
